#!/usr/bin/env python
DESC="""EPP used to create running notes from the workset generation """

from argparse import ArgumentParser
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.epp import attach_file, EppLogger
from genologics.entities import Process, Project
from datetime import datetime

import json
import sys
import os


def main(lims, args):

    p=Process(lims, id=args.pid)
    log=[]
    datamap={}
    wsname=None
    username="{0} {1}".format(p.technician.first_name, p.technician.last_name)
    user_email=p.technician.email
    for art in p.all_inputs():
        if len(art.samples)!=1:
            log.append("Warning : artifact {0} has more than one sample".format(art.id))
        for sample in art.samples:
           #take care of lamda DNA
           if sample.project:
                if sample.project.id not in datamap:
                    datamap[sample.project.id]=[sample.name]
                else:
                    datamap[sample.project.id].append(sample.name)

    for art in p.all_outputs():
        try:
            wsname=art.location[0].name
            break
        except:
            pass

    now=datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    for pid in datamap:
        pj=Project(lims, id=pid)
        running_notes=json.loads(pj.udf['Running Notes'])
        if len(datamap[pid]) > 1:
            rnt="{0} samples planned for {1}".format(len(datamap[pid]), wsname)
        else:
            rnt="{0} sample planned for {1}".format(len(datamap[pid]), wsname)

        running_notes[now]={"note": rnt, "user" : username, "email":user_email, "category":"Workset"}

        pj.udf['Running Notes']=json.dumps(running_notes)
        pj.put()
        log.append("Updated project {0} : {1}, {2} samples in this workset".format(pid,pj.name, len(datamap[pid])))


 
    with open("EPP_Notes.log", "w") as flog:
        flog.write("\n".join(log))
    for out in p.all_outputs():
        #attach the log file
        if out.name=="RNotes Log":
            attach_file(os.path.join(os.getcwd(), "EPP_Notes.log"), out)

    sys.stderr.write("Updated {0} projects successfully".format(len(datamap.keys())))

if __name__=="__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()
    main(lims, args)



#!/usr/bin/env python
DESC="""EPP script to copy the "Comments" field to the projects running notes on process termination

Denis Moreno, Science for Life Laboratory, Stockholm, Sweden
""" 
from argparse import ArgumentParser
from genologics.entities import *
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.epp import attach_file, EppLogger

import datetime
import logging
import os
import sys
import json


def categorization(process_name):
    decision={
    "Automated Quant-iT QC (Library Validation) 4.0" : "Workset",
    "Library Pooling (MiSeq) 4.0" : "",
    "Aggregate QC (Library Validation) 4.0" : "Workset",
    "Sample Placement (Size Selection)" : "",
    "Illumina Sequencing (HiSeq X) 1.0" : "Flowcell",
    "Ligate 5' adapters (TruSeq small RNA) 1.0" : "Workset",
    "End repair, A-tailing and adapter ligation (Nextera) 4.0" : "Workset",
    "Library Normalization (Illumina SBS) 4.0" : "",
    "Bioanalyzer QC (DNA) 4.0" : "Workset",
    "Pre-Pooling (MiSeq) 4.0" : "",
    "ThruPlex library amplification" : "Workset",
    "Applications Pre-Pooling" : "",
    "Sort HiSeq Samples (HiSeq) 4.0" : "",
    "mRNA Purification, Fragmentation & cDNA synthesis (TruSeq RNA) 4.0" : "Workset",
    "HT-End repair, A-tailing and adapter ligation (TruSeq RNA) 4.0" : "Workset",
    "Size Selection (Pippin)" : "Workset",
    "End Repair, A-Tailing and Adapter Ligation (SS XT) 4.0" : "Workset",
    "Qubit QC (Library Validation) 4.0" : "Workset",
    "Library Pooling (Finished Libraries) 4.0" : "",
    "MiSeq Run (MiSeq) 4.0" : "Flowcell",
    "Automated Quant-iT QC (DNA) 4.0" : "Workset",
    "Bioanalyzer QC (Library Validation) 4.0" : "Workset",
    "Setup Workset/Plate" : "Workset",
    "Bioanalyzer Fragmentation QC (TruSeq DNA) 4.0" : "Workset",
    "Ligate 3' adapters (TruSeq small RNA) 1.0" : "Workset",
    "Qubit QC (DNA) 4.0" : "Workset",
    "Sort MiSeq Samples (MiSeq) 4.0" : "",
    "Fragment DNA (ThruPlex)" : "Workset",
    "Volume Measurement QC" : "Workset",
    "Project Summary 1.3" : "",
    "Fragment DNA (TruSeq DNA) 4.0" : "Workset",
    "NeoPrep Library QC v1.0" : "Workset",
    "Library Normalization (MiSeq) 4.0" : "",
    "Aliquot Samples for Caliper/Bioanalyzer" : "Workset",
    "qPCR QC (Library Validation) 4.0" : "Workset",
    "Applications Indexing" : "Workset",
    "CA Purification" : "Workset",
    "Size Selection (Caliper XT) 1.0" : "Workset",
    "ThruPlex template preparation and synthesis" : "Workset",
    "Enrich DNA fragments (TruSeq RNA) 4.0" : "Workset",
    "Quant-iT QC (RNA) 4.0" : "Workset",
    "End repair, size selection, A-tailing and adapter ligation (TruSeq PCR-free DNA) 4.0" : "Workset",
    "CaliperGX QC (DNA)" : "Workset",
    "Hybridize Library  (SS XT) 4.0" : "",
    "Sort HiSeq X Samples (HiSeq X) 1.0" : "",
    "Amplify Adapter-Ligated Library (SS XT) 4.0" : "",
    "Aliquot Libraries for Hybridization (SS XT)" : "",
    "Denature, Dilute and Load Sample (MiSeq) 4.0" : "Flowcell",
    "Cluster Generation (HiSeq X) 1.0" : "Flowcell",
    "Cluster Generation (Illumina SBS) 4.0" : "Flowcell",
    "Library Pooling (Illumina SBS) 4.0" : "",
    "Bioanalyzer QC (RNA) 4.0" : "Workset",
    "Pre-Pooling (Illumina SBS) 4.0" : "",
    "Reverse Transcribe (TruSeq small RNA) 1.0" : "Workset",
    "RiboZero depletion" : "Workset",
    "Amplify Captured Libraries to Add Index Tags (SS XT) 4.0" : "",
    "Library Pooling (TruSeq Small RNA) 1.0" : "",
    "Linear DNA digestion, Circularized DNA shearing and Streptavidin Bead Binding" : "Workset",
    "CaliperGX QC (RNA)" : "Workset",
    "Bcl Conversion & Demultiplexing (Illumina SBS) 4.0" : "Bioinformatics",
    "Qubit QC (Dilution Validation) 4.0" : "Workset",
    "MinElute Purification" : "",
    "Quant-iT QC (DNA) 4.0" : "Workset",
    "Aggregate QC (RNA) 4.0" : "Workset",
    "Customer Gel QC" : "Workset",
    "Quant-iT QC (Library Validation) 4.0" : "Workset",
    "Circularization" : "Workset",
    "Qubit QC (RNA) 4.0" : "Workset",
    "Illumina Sequencing (Illumina SBS) 4.0" : "Flowcell",
    "Purification (ThruPlex)" : "Workset",
    "Aliquot Samples for Qubit/Bioanalyzer" : "Workset",
    "Shear DNA (SS XT) 4.0" : "Workset",
    "Tagmentation, Strand displacement and AMPure purification" : "Workset",
    "Fragmentation & cDNA synthesis (TruSeq RNA) 4.0" : "Workset",
    "End repair, A-tailing and adapter ligation (TruSeq RNA) 4.0" : "Workset",
    "Capture And Wash (SS XT) 4.0" : "Workset",
    "Amplify by PCR and Add Index Tags (TruSeq small RNA) 1.0" : "Workset",
    "Size Selection (Robocut)" : "Workset",
    "Aggregate QC (DNA) 4.0" : "Workset",
    "Aliquot Libraries for Pooling (Small RNA)" : ""}

    return decision[process_name]



def main(lims, args):
    

    comment=False

    noteobj={}
    pro=Process(lims, id=args.pid)
    if 'Comments' in pro.udf and pro.udf['Comments'] is not '':
        key=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        noteobj[key]={}
        note="Comment from {0} ({1}) : \n{2}".format(pro.type.name, '[LIMS](https://genologics.scilifelab.se/clarity/work-details/{0})'.format(pro.id.split('-')[1]), pro.udf['Comments'])
        noteobj[key]['note']=note
        noteobj[key]['user']="{0} {1}".format(pro.technician.first_name,pro.technician.last_name)
        noteobj[key]['email']=pro.technician.email
        noteobj[key]['category']=categorization(pro.type.name)

        #find the correct projects.
        samples=set()
        projects=set()
        for inp in pro.all_inputs():
            #bitwise or to add inp.samples to samplesas a set
            samples |= set(inp.samples)
        for sam in samples:
            projects.add(sam.project)

        for proj in projects:
            if 'Running Notes' in proj.udf:
                existing_notes=json.loads(proj.udf['Running Notes'])
                for key in noteobj:
                    existing_notes[key]=noteobj[key]
                proj.udf['Running Notes']=json.dumps(existing_notes)
            else:
                proj.udf['Running Notes']=json.dumps(noteobj) 
            proj.put()
            
if __name__=="__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()

    main(lims, args)



#!/usr/bin/env python
DESC = """This EPP script reads demultiplex end undemultiplexed yields from file
system and then does the following

1)  Sets the output artifact udfs based demultiplex resultfile.
 
2)  Generated qc tresholds based on sequencing run setings. These can also be 
    set by user.
 
3)  Loggs the tresholds to a qc log file

4)  Sets the output artifact qc-flaggs based on the threshold

5)  Warns if anny unexpected index has high yield ore if a lane has high yield 
    of un expected indexes

6)  Loads a result file with demultiplex end undemultiplexed yields. This should
    be checked if warnings are given.

Reads from:
    --files--
    Demultiplex_Stats.htm                           in mfs file system
    Undemultiplexed_stats.metrics                   in mfs file system

Logging:
    The script outputs a regular log file with regular execution information.

Written by Maya Brandi 
"""

import os
import sys
import logging
import glob
import csv
import numpy as np

from argparse import ArgumentParser
from genologics.lims import Lims
from genologics.config import BASEURI, USERNAME, PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import set_field
from scilifelab_parsers.qc.qc import FlowcellRunMetricsParser
#from qc_parsers import FlowcellRunMetricsParser

class RunQC():
    def __init__(self, process):
        ## Processes, artifacts and udfs 
        self.process = process
        self.input_pools = process.all_inputs()
        self.seq_run = None
        self.run_udfs = {}
        self.user_def_tresh = dict(self.process.udf.items())

        ##  Stuff for logging 
        self.qc_log_file = None
        self.nr_lane_samps_updat = 0
        self.nr_lane_samps_tot = 0
        self.abstract = []
        self.QC_fail = []
        self.high_index_yield = []
        self.high_lane_yield = []
        self.html_file_error = False ## Never used????

        ##  Demultiplexing result files
        self.file_path = None
        self.dem_stat = None
        self.undem_stat = None 

        ##  Other variables
        self.single = True
        self.read_length = None
        self.Q30_treshold = None

    def make_qc_log_file(self, qc_log_file):
        """File to logg qc tresholds."""
        self.qc_log_file = open(qc_log_file, 'a')

    def get_run_info(self):
        cont_name = self._get_container()
        self._get_run(cont_name)
        self.run_udfs = dict(self.seq_run.udf.items())
        self._get_run_id()
        self._get_cycles()
        self._get_file_path(cont_name)
        self._get_demultiplex_files()
        self._get_threshold_Q30()

    def _get_container(self):
        """Need container name to fetch parrent sequencing process"""
        try:
            return self.process.all_inputs()[0].location[0].name
        except:
            sys.exit('Could not find container name.')
    
    def _get_run(self, cont_name):
        """Getting parrent sequencing process and process type"""
        miseq = lims.get_processes(udf = {'Reagent Cartridge ID' : cont_name},
                                type = 'MiSeq Run (MiSeq) 4.0')
        hiseq = lims.get_processes(udf = {'Flow Cell ID' : cont_name},
                                type = 'Illumina Sequencing (Illumina SBS) 4.0')
        hiseq_X10 = lims.get_processes(udf = {'Flow Cell ID' : cont_name},
                                type = 'Illumina Sequencing (HiSeq X) 1.0')
        if miseq:
            self.seq_run = miseq[0]
            self.run_type = 'MiSeq'
        elif hiseq:
            self.seq_run = hiseq[0]
            try:
                self.run_type = self.seq_run.udf['Flow Cell Version']
            except:
                sys.exit("Missing field 'Flow Cell Version' in sequencing process")
        elif hiseq_X10:
            self.seq_run = hiseq_X10[0]
            self.run_type = 'HiSeqX10'
        else:
            sys.exit("run not found")

    def _get_run_id(self):
        if self.run_udfs.has_key('Run ID'):
            self.process.udf['Run ID'] = self.run_udfs['Run ID']
            set_field(self.process)

    def _get_cycles(self):
        """To find out if it was a single end or paired end run."""
        if self.run_udfs.has_key('Read 1 Cycles'):
            self.read_length = self.run_udfs['Read 1 Cycles']
        else:
            sys.exit("Could not get 'Read 1 Cycles' from the sequencing step.")
        if self.run_udfs.has_key('Read 2 Cycles'):
            self.single = False

    def _get_file_path(self, cont_name):
        """File path for HiSeq, Miseq and Xten"""
        if self.run_type == 'MiSeq':
            path_id = self.run_udfs['Flow Cell ID']
            data_folder = 'miseq_data'
        elif self.run_type == 'HiSeqX10':
            path_id = cont_name
            data_folder = 'HiSeq_X_data'
        else:
            data_folder = 'hiseq_data'
            path_id = cont_name
        try:
            self.file_path = glob.glob(("/srv/mfs/{0}/*{1}/Unaligned/"
                           "Basecall_Stats_*/".format(data_folder, path_id)))[0]
        except:
            sys.exit("Failed to get file path")

    def _get_demultiplex_files(self):
        """ Files are read from the file msf system. Path hard coded."""
        FRMP = FlowcellRunMetricsParser()
        try:
            fp_dem = self.file_path + 'Demultiplex_Stats.htm'
            self.dem_stat = FRMP.parse_demultiplex_stats_htm(fp_dem)
            logging.info("Parsed file {0}".format(fp_dem))
        except:
            sys.exit("Failed to find or parse Demultiplex_Stats.htm.")
        try:
            fp_und = self.file_path + 'Undemultiplexed_stats.metrics'
            self.undem_stat = FRMP.parse_undemultiplexed_barcode_metrics(fp_und)
            logging.info("Parsed file {0}".format(fp_und))
        except:
            sys.exit("Failed to find or parse Undemultiplexed_stats.metrics")

    def _get_threshold_Q30(self):
        if self.user_def_tresh.has_key('Threshold for % bases >= Q30'):
            self.Q30_treshold = self.user_def_tresh['Threshold for % bases >= Q30']
            qc_logg = ("THRESHOLD FOR %Q30 was set by user to {0}.".format(
                        self.Q30_treshold))
            print >> self.qc_log_file, qc_logg
        else:
            warning = ("Un recognized read length: {0}. Report this to "
                "developers! set Threshold for % bases >= Q30 if you want to "
                "run bcl conversion and demultiplexing anyway.".format(self.read_length))
            if self.run_type == 'MiSeq':
                if self.read_length < 101:
                    Q30_threshold = 80
                elif self.read_length == 101:
                    Q30_threshold = 75
                elif self.read_length == 151:
                    Q30_threshold = 70
                elif self.read_length >= 251:
                    Q30_threshold = 60
                else:
                    sys.exit(warning)
            else:
                if self.read_length == 51:
                    Q30_threshold = 85
                elif self.read_length == 101:
                    Q30_threshold = 80
                elif self.read_length == 126:
                    Q30_threshold = 80
                elif self.read_length == 151:
                    Q30_threshold = 75
                else:
                    sys.exit(warning)
            qc_logg = ("THRESHOLD FOR %Q30 was set to {0}. Value based on read"
                       " length: {1}, and run type {2}.".format(Q30_threshold, 
                                               self.read_length, self.run_type))
            print >> self.qc_log_file, qc_logg
            self.Q30_treshold = Q30_threshold

    def run_QC(self):
        for pool in self.input_pools:
            outarts_per_lane = self.process.outputs_per_input(pool.id, ResultFile = True)
            lane_number = '1' if self.run_type == 'MiSeq' else pool.location[1][0]
            LQC = LaneQC(lane_number, outarts_per_lane, self.run_type, 
                         self.undem_stat, self.dem_stat, self.single, 
                         self.Q30_treshold, self.qc_log_file, self.user_def_tresh, self.read_length)
            LQC.set_and_log_tresholds()
            LQC.lane_QC()
            self.nr_lane_samps_tot += LQC.nr_lane_samps
            self.nr_lane_samps_updat += LQC.nr_samps_updat
            self.QC_fail += LQC.QC_fail
            if LQC.high_lane_yield:
                self.high_lane_yield.append(LQC.lane)
            if LQC.high_index_yield:
                self.high_index_yield.append(LQC.lane)

        ## moove this part to logging???-->>
        if self.high_index_yield or self.high_lane_yield:
            warn = "WARNING: "
            if self.high_index_yield:
                self.high_index_yield = ', '.join(list(set(self.high_index_yield)))
                warn = ("{0} High yield of unexpected index on lane(s): {1} ."
                        "".format(warn, self.high_index_yield))
            if self.high_lane_yield:
                self.high_lane_yield = ', '.join(list(set(self.high_lane_yield)))
                warn = ("{0} High total yield of unexpected index on lane(s): "
                        "{1}.".format(warn, self.high_lane_yield))
            warn = warn + "Please check the Metrics file!"
            self.abstract.insert(0, warn)

    def make_demultiplexed_counts_file(self, demuxfile):
        """Reformats the content of the demultiplex and undemultiplexed files
        to be more easy to read."""

        demuxfile = demuxfile + '.csv'
        keys = ['Project', 'Sample ID', 'Lane', '# Reads', 'Index',
                                    'Index name', '% of >= Q30 Bases (PF)']
        toCSV = []
        for pool in self.input_pools:
            if self.run_type == 'MiSeq':
                lane = '1'
            else:
                lane = pool.location[1][0]
            for row in self.dem_stat['Barcode_lane_statistics']:
                if row['Lane'] == lane:
                    row_dict = dict([(x, row[x]) for x in keys if x in row])
                    row_dict['Index name'] = ''
                    toCSV.append(row_dict)
            if lane in self.undem_stat.keys():
                undet_per_lane = self.undem_stat[lane]['undemultiplexed_barcodes']
                nr_undet = len(undet_per_lane['count'])
                for row in range(nr_undet):
                    row_dict = dict([(x, '') for x in keys])
                    row_dict['# Reads'] = undet_per_lane['count'][row]
                    row_dict['Index'] = undet_per_lane['sequence'][row]
                    row_dict['Index name'] = undet_per_lane['index_name'][row]
                    row_dict['Lane'] = undet_per_lane['lane'][row]
                    toCSV.append(row_dict)
        try:
            f = open(demuxfile, 'w')
            dict_writer = csv.DictWriter(f, keys, dialect='excel')
            dict_writer.writer.writerow(keys)
            dict_writer.writerows(toCSV)
            f.close
            self.abstract.append("INFO: A Metrics file has been created with "
                      "demultiplexed and undemultiplexed counts for debugging.")
        except:
            self.abstract.append("WARNING: Could not generate a Metrics file "
                               "with demultiplexed and undemultiplexed counts.")

    def logging(self):
        """Collects and prints logging info."""
        self.abstract.append("INFO: QC-data found and QC-flags uploaded for {0}"
              " out of {1} analytes.".format(self.nr_lane_samps_updat,
                                                        self.nr_lane_samps_tot))
        if self.QC_fail:
            self.abstract.append('Failed to make qc for samples: {0}'.format(
                ', '.join(list(set(self.QC_fail)))))
        if 'WARNING' in ' '.join(self.abstract):
            sys.exit(' '.join(self.abstract))
        else:
            print >> sys.stderr, ' '.join(self.abstract)


class LaneQC():
    def __init__(self, lane_number ,out_arts, run_type, undem_stat, dem_stat, 
                 single, Q30_treshold, qc_log_file, user_def_tresholds, read_length):
        ##  Output artifacts and user defined tresholds
        self.out_arts = out_arts
        self.user_def_tresh = user_def_tresholds

        ##  Info from files in file system
        self.counts = undem_stat[lane_number]['undemultiplexed_barcodes']['count']
        self.BLS = dem_stat['Barcode_lane_statistics']

        ##  Tresholds
        self.exp_lane_clust = None
        self.reads_threshold = None 
        self.un_exp_lane = None 
        self.thres_un_exp_ind = None 
        self.Q30_treshold = Q30_treshold 

        ##  Stuff for logging 
        self.qc_log_file = qc_log_file
        self.high_lane_yield = False
        self.high_index_yield = False
        self.nr_samps_updat = 0
        self.html_file_error = False
        self.QC_fail = []

        ##  Other variables
        self.single = single
        self.read_length = read_length
        self.run_type = run_type
        self.lane  = lane_number
        self.nr_lane_samps = len(out_arts)

    def set_and_log_tresholds(self):
        """Generating tresholds and writing the tresholds to log file."""
        print >> self.qc_log_file, ''
        print >> self.qc_log_file, 'TRESHOLDS - LANE {0}:'.format(self.lane)
        self._get_exp_lane_and_ind_clust()
        self._set_reads_threshold()
        self._set_tresh_un_exp_lane()
        self._set_tresh_un_exp_ind()

    def _get_exp_lane_and_ind_clust(self):
        """The expected number of lane clusters depends on run type and run mode.
        The expected number of clusters per index (sample) is then:
        (expected nr clusters on the lane)/(nr indexes on the lane)"""
        if self.run_type == 'MiSeq':
            if self.read_length in [76, 301]:  
                self.exp_lane_clust = 18000000
            else:                               
                self.exp_lane_clust = 10000000
        elif self.run_type in ['HiSeq Rapid Flow Cell v1','HiSeq Rapid Flow Cell v2', "TruSeq Rapid Flow Cell v2", "TruSeq Rapid Flow Cell v1"] :
            self.exp_lane_clust = 114000000
        elif self.run_type == 'HiSeq Flow Cell v3':
            self.exp_lane_clust = 143000000
        elif self.run_type == 'HiSeq Flow Cell v4':
            self.exp_lane_clust = 188000000
        elif self.run_type == 'HiSeqX10':
            self.exp_lane_clust = 250000000
        else:
            sys.exit('Unrecognized run type: {0}. Report to developer! Set '
                    'Threshold for # Reads if you want to run bcl conversion '
                    'and demultiplexing again'.format(self.run_type))
        self.exp_samp_clust = np.true_divide(self.exp_lane_clust, self.nr_lane_samps)

    def _set_reads_threshold(self):
        """Treshold for nr reads per index: 
            exp_samp_clust*0.5"""

        if self.user_def_tresh.has_key('Threshold for # Reads'):
            self.reads_threshold = self.user_def_tresh['Threshold for # Reads']
            qc_logg = "Index yield - expected index: {0}".format(self.reads_threshold)
            print >> self.qc_log_file , qc_logg
        else:
            exp_samp_clust = np.true_divide(self.exp_lane_clust, self.nr_lane_samps)
            self.reads_threshold = int(self.exp_samp_clust*0.5)
            qc_logg = ("Index yield - expected index: {0}. Value based on nr of "
                        "sampels in the lane: {1}, and run type {2}.".format(
                        self.reads_threshold, self.nr_lane_samps, self.run_type))
            print >> self.qc_log_file , qc_logg

    def _set_tresh_un_exp_lane(self):
        """Treshold for un expected indexes on a hole lane:
            If single end: exp_lane_clust*0.05
            If paired end: exp_lane_clust*0.1"""

        if self.single: 
            self.un_exp_lane = int(self.exp_lane_clust*0.05) 
        else:
            self.un_exp_lane = int(self.exp_lane_clust*0.1)
        qc_logg = ("Lane yield - un expected index: {0}. Value based on run "
                 "type {1}, and run setings: {2}".format(self.un_exp_lane, 
                 self.run_type, 'Single End' if self.single else 'Paired End'))
        print >> self.qc_log_file, qc_logg

    def _set_tresh_un_exp_ind(self):
        """Threshold for un expected index:
            exp_samp_clust*0.1"""

        if self.user_def_tresh.has_key('Threshold for Undemultiplexed Index Yield'):
            self.thres_un_exp_ind = self.user_def_tresh['Threshold for Undemultiplexed Index Yield']
            qc_logg = ("Index yield - un expected index: {0}. Value set by user."
                     "".format(self.thres_un_exp_ind))
            print >> self.qc_log_file, qc_logg
        else:
            self.thres_un_exp_ind = int(self.exp_samp_clust*0.1)
            qc_logg = ("Index yield - un expected index: {0}. Value set to 10% "
                    "of expected index yield".format(self.thres_un_exp_ind))
            print >> self.qc_log_file, qc_logg

    def lane_QC(self):
        for target_file in self.out_arts:
            samp_name = target_file.samples[0].name
            for lane_samp in self.BLS:
                if self.lane == lane_samp['Lane']:
                    samp = lane_samp['Sample ID']
                    if samp == samp_name:
                        IQC = IndexQC(target_file, lane_samp)
                        IQC.set_target_file_udfs()
                        IQC.set_read_pairs(self.single)
                        try:
                            IQC.lane_index_QC(self.reads_threshold, self.Q30_treshold)
                            if IQC.html_file_error:
                                self.html_file_error = IQC.html_file_error
                            set_field(IQC.t_file)
                            self.nr_samps_updat +=1
                        except:
                            self.QC_fail.append(samp)
        self._check_un_exp_lane_yield()
        for index_count in self.counts:
            self._check_un_exp_ind_yield(index_count)

    def _check_un_exp_lane_yield(self):
        unexp_lane_yield = sum([int(x) for x in self.counts])
        if unexp_lane_yield > self.un_exp_lane:
            self.high_lane_yield = True

    def _check_un_exp_ind_yield(self, index_count):
        if int(index_count) > self.thres_un_exp_ind:
            self.high_index_yield = True


class IndexQC():
    def __init__(self, t_file, stats):
        self.t_file = t_file
        self.stats = stats
        self.samp_udfs = {}
        self.html_file_error = False

    def set_target_file_udfs(self):
        """ Populates the target file udfs (run lane index resolution) with 
        Barcode lane statistics fetched from demultiplexed stats file"""

        self.samp_udfs = {'%PF' : self.stats['% PF'],
            '% One Mismatch Reads (Index)' : self.stats['% One Mismatch Reads (Index)'],
            '% of Raw Clusters Per Lane' : self.stats['% of raw clusters per lane'],
            'Ave Q Score' : self.stats['Mean Quality Score (PF)'],
            '% Perfect Index Read' : self.stats['% Perfect Index Reads']}
        self._make_float()
        self._set_Yield_Mbases()
        self._set_Q30()
        self._set_reads()

    def _make_float(self):
        for key, val in self.samp_udfs.items():
            try:
                self.t_file.udf[key] = float(val)
            except:
                self.html_file_error = True

    def _set_Yield_Mbases(self):
        try:
            yMb = self.stats['Yield (Mbases)'].replace(',','')
            self.t_file.udf['Yield PF (Gb)'] = np.true_divide(float(yMb), 1000)
        except:
            self.html_file_error = True

    def _set_Q30(self):
        if not dict(self.t_file.udf.items()).has_key('% Bases >=Q30'):
            self.t_file.udf['% Bases >=Q30'] = self.stats['% of >= Q30 Bases (PF)']

    def _set_reads(self):
        if not dict(self.t_file.udf.items()).has_key('# Reads'):
            try:
                self.t_file.udf['# Reads'] = float(self.stats['# Reads'].replace(',',''))
            except:
                self.html_file_error = True

    def set_read_pairs(self, single):
        if single and self.t_file.udf['# Reads']:
            self.t_file.udf['# Read Pairs'] = int(self.t_file.udf['# Reads'])
        elif self.t_file.udf['# Reads']:
            reads = float(self.t_file.udf['# Reads'])
            self.t_file.udf['# Read Pairs'] = np.true_divide(reads, 2)

    def lane_index_QC(self, reads_threshold, Q30_treshold):
        """Index-lane level QC based on derieved QC-tresholds. OBS: Fetches info
        from target file udf if they are already set. Otherwise from file 
        system!!! This is to take into account yield and quality after quality 
        filtering if performed."""

        Q30 = float(self.stats['% of >= Q30 Bases (PF)'])
        QC2 = (Q30 >= Q30_treshold)
        QC3 = (self.t_file.udf['# Read Pairs'] >= reads_threshold)
        if QC2 and QC3:
            self.t_file.udf['Include reads'] = 'YES'
            self.t_file.qc_flag = 'PASSED'
        else:
            self.t_file.udf['Include reads'] = 'NO'
            self.t_file.qc_flag = 'FAILED'

def main(lims, pid, epp_logger, demuxfile, qc_log_file):
    process = Process(lims,id = pid)
    RQC = RunQC(process)
    RQC.make_qc_log_file(qc_log_file)
    RQC.get_run_info()
    RQC.run_QC()
    RQC.make_demultiplexed_counts_file(demuxfile)
    RQC.logging()

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = None , dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))
    parser.add_argument('--file', dest = 'file', default = 'demux',
                        help=('File path to demultiplexed metrics file'))
    parser.add_argument('--qc_log_file', dest = 'qc_log_file', 
                        default = 'qc_log_file',
                        help=('File path to qc logfile file'))
    args = parser.parse_args()
    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, epp_logger, args.file, args.qc_log_file)



#!/usr/bin/env python
DESC = """EPP script to copy user defined field from any process to associated 
project/projects in Clarity LIMS. If the specifyed process handles many artifacts 
associated to different projects, all these projects will get the specifyed udf.
 
Can be executed in the background or triggered by a user pressing a "blue button".

The script can output two different logs, where the status_changelog 
contains notes with the technician, the date and changed status for each 
copied status. The regular log file contains regular execution information.

Written by Maya Brandi 
"""
import os
import sys
import logging

from argparse import ArgumentParser

from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import CopyField

def main(lims, args, epp_logger):
    d_elts = []
    no_updated = 0
    incorrect_udfs = 0
    project_names = ''
    source_udfs = args.source_udf
    dest_udfs = args.dest_udf
    s_elt = Process(lims,id = args.pid)
    analytes, inf = s_elt.analytes()

    for analyte in analytes:
        for samp in analyte.samples:
            d_elts.append(samp.project)
    d_elts = list(set(d_elts))

    if args.status_changelog:
        epp_logger.prepend_old_log(args.status_changelog)

    if not dest_udfs:
        dest_udfs = source_udfs
    elif len(dest_udfs) != len(source_udfs):
        logging.error("source_udfs and dest_udfs lists of arguments are uneven.")
        sys.exit(-1)
    
    for d_elt in d_elts:
        project_names = ' '.join([project_names, d_elt.name])
        for i in range(len(source_udfs)):
            source_udf = source_udfs[i]
            dest_udf = dest_udfs[i]
            with open(args.status_changelog, 'a') as changelog_f:
                if source_udf in s_elt.udf:
                    copy_sesion = CopyField(s_elt, d_elt, source_udf, dest_udf)
                    test = copy_sesion.copy_udf(changelog_f)
                    if test:
                        no_updated = no_updated + 1
                else:
                    logging.warning(("Udf: {1} in Process {0} is undefined/blank, exiting").format(s_elt.id, source_udf))
                    incorrect_udfs = incorrect_udfs + 1

    if incorrect_udfs > 0:
        warn = "Failed to update %s udf(s) due to missing/wrong source udf info." %incorrect_udfs
    else:
        warn = ''

    d = {'up': no_updated,
         'ap': len(d_elts),
         'w' : warn,
         'pr': project_names}

    abstract = ("Updated {up} udf(s). Handeled project(s): {pr} {w}").format(**d)
    print >> sys.stderr, abstract

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))
    parser.add_argument('-s', '--source_udf', type=str, default=None, nargs='*',
                        help=('Name(s) of the source user defined field(s) '
                               'that will be copied. One or many udf-names '
                               'can be given.'))
    parser.add_argument('-d', '--dest_udf', type=str, default=None, nargs='*',
                        help=('Name(s) of the destination user defined '
                              'field(s) that will be written to. This '
                              'argument is optional, if left empty '
                              'the source_udf argument is used instead. '
                              'Zero or many udf-names can be given. If '
                              'more than zero, the numer of udfs needs ' 
                              'to be the same as number of source_udfs'))
    parser.add_argument('-c', '--status_changelog',
                        help=('File name for status changelog file, for '
                              'concise information on who, what and when '
                              'for status change events. '
                              'Prepends the old changelog file by default.'))

    args = parser.parse_args()

    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)




#!/usr/bin/env python
DESC = """EPP script for Quant-iT measurements to generate a driver file (.csv) 
which include the Sample names and their positions in the working plate.

Reads from:
    --Lims fields--
    "location"      field of input analytes to the process

Writes to:
    --file--
    "Driver File"   shared result file

Logging:
The script outputs a regular log file with regular execution information.

Written by Maya Brandi 
"""

import os
import sys
import logging

from argparse import ArgumentParser
from requests import HTTPError
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import set_field
from genologics.epp import ReadResultFiles

class QuantitDriverFile():
    def __init__(self, process, drivf):
        self.udfs = dict(process.udf.items())
        self.drivf = drivf

    def make_location_dict(self, io_filtered):
        """Loops through the input-output map and formates the 
        well location info into the driver file formate: 
        row,col,,sample_name"""
        location_dict = {}
        for input, output in io_filtered:
            well = output['uri'].location[1]
            sample = input['uri'].name
            row, col = well.split(':')
            location_dict[well] = ','.join([row, col,'', sample])
        return location_dict

    def make_file(self, location_dict):
        """Writes the formated well location info into a driver 
        file sorted by row and col."""
        keylist = location_dict.keys()
        keylist.sort()
        f = open(self.drivf, 'a')
        print >> f , 'Row,Column,*Target Name,*Sample Name'
        for key in keylist:
            print >> f ,location_dict[key]
        f.close()

def main(lims, pid, drivf ,epp_logger):
    process = Process(lims,id = pid)
    QiT = QuantitDriverFile(process, drivf)
    io = process.input_output_maps
    io_filtered = filter(lambda (x,y): y['output-generation-type']=='PerInput', io)
    io_filtered = filter(lambda (x,y): y['output-type']=='ResultFile', io_filtered)
    location_dict = QiT.make_location_dict(io_filtered)
    QiT.make_file(location_dict)

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = None , dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))
    parser.add_argument('--drivf', dest = 'drivf', 
                        default = 'QuantiT_driver_file_exported_from_LIMS.csv',
                        help=('File name for Driver file to be generated'))

    args = parser.parse_args()
    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, args.drivf, epp_logger)




#!/usr/bin/env python
import sys
from argparse import ArgumentParser
import subprocess
import logging
import datetime
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.lims import Lims
from genologics.epp import EppLogger
from genologics.entities import Process

def makeContainerBarcode(plateid,copies=1):
    """ Construct label with container id as human readable and barcode """
    lines = []
    lines.append("^XA") #start of label
    # download and store format, name of format, 
    # end of field data (FS = field stop)
    lines.append("^DFFORMAT^FS") 
    lines.append("^LH0,0") # label home position (label home = LH)
    # AF = assign font F, field number 1 (FN1), 
    # print text at position field origin (FO) rel. to home
    lines.append("^FO360,20^AFN 78,39^FN1^FS")
    # BC=barcode 128, field number 2, Normal orientation, 
    # height 70, no interpreation line. 
    lines.append("^FO70,5^BCN,70,N,N^FN2^FS")
    lines.append("^XZ") #end format

    for copy in xrange(copies):
        lines.append("^XA") #start of label format
        lines.append("^XFFORMAT^FS") #label home position
        lines.append("^FN1^FD"+plateid+"^FS") #this is readable
        lines.append("^FN2^FD"+plateid+"^FS") #this is also readable
        lines.append("^XZ")
    return lines

def makeContainerNameBarcode(plate_name,copies=1):
    """ Constrcut label with container name as human readable """
    lines = []
    lines.append("^XA") #start of label
    # download and store format, name of format, 
    # end of field data (FS = field stop)
    lines.append("^DFFORMAT^FS") 
    lines.append("^LH0,0") # label home position (label home = LH)
    # AF = assign font F, field number 1 (FN1), 
    # print text at position field origin (FO) rel. to home
    if len(plate_name)>21:
        # Use smaller font, fits 28 chars
        lines.append("^FO20,30^ADN 54,30^FN1^FS")
    else:
        # Use larger font, fits 21 chars
        lines.append("^FO20,20^AFN 78,39^FN1^FS")

    lines.append("^XZ") #end format

    for copy in xrange(copies):
        lines.append("^XA") #start of label format
        lines.append("^XFFORMAT^FS") #label home position
        lines.append("^FN1^FD"+plate_name+"^FS") #this is readable
        lines.append("^XZ")
    return lines

def makeOperatorAndDateBarcode(operator,date,copies=1):
    """ Construct label with operator name and date in human readable format"""
    lines = []
    lines.append("^XA") #start of label
    # Download and store format, name of format, 
    # end of field data (FS = field stop)
    lines.append("^DFFORMAT^FS") 
    lines.append("^LH0,0") # label home position (label home = LH)
    # AF = assign font F, field number 1 (FN1), 
    # print text at position field origin (FO) rel. to home
    lines.append("^FO460,30^ADN,36,20^FN1^FS")
    lines.append("^FO20, 30^ADN,36,20^FN2^FS")
    lines.append("^XZ") #end format

    if len(operator)>19:
        operator = operator[:19] # If string is longer, it would cover the date
    for copy in xrange(copies):
        lines.append("^XA") #start of label format
        lines.append("^XFFORMAT^FS") #label home position
        lines.append("^FN1^FD"+date+"^FS") #this is readable
        lines.append("^FN2^FD"+operator+"^FS") #this is also readable
        lines.append("^XZ")
    return lines

def makeProcessNameBarcode(process_name,copies=1):
    """ Constrcut label with process name as human readable """
    lines = []
    lines.append("^XA") #start of label
    # download and store format, name of format, 
    # end of field data (FS = field stop)
    lines.append("^DFFORMAT^FS") 
    lines.append("^LH0,0") # label home position (label home = LH)
    # AF = assign font F, field number 1 (FN1), 
    # print text at position field origin (FO) rel. to home
    if len(process_name)>21:
        # Use smaller font, fits 28 chars
        lines.append("^FO20,30^ADN 54,30^FN1^FS")
    else:
        # Use larger font, fits 21 chars
        lines.append("^FO20,20^AFN 78,39^FN1^FS")

    lines.append("^XZ") #end format

    for copy in xrange(copies):
        lines.append("^XA") #start of label format
        lines.append("^XFFORMAT^FS") #label home position
        lines.append("^FN1^FD"+process_name+"^FS") #this is readable
        lines.append("^XZ")
    return lines
    
def getArgs():
    desc = (" Print barcodes on zebra barcode printer, "
            " different label types available. Information "
            " is fetched from Clarity LIMS.")
    parser = ArgumentParser(description=desc)
    parser.add_argument('--container_id', action='store_true',
                        help=('Print output container id label in both '
                        'barcode format and human readable.'))
    parser.add_argument('--operator_and_date',action='store_true',
                        help=('Print label with both operator '
                              'and todays date.'))
    parser.add_argument('--container_name',action='store_true',
                        help=('Print label with human readable'
                              'container name (user defined)'))
    parser.add_argument('--process_name', action='store_true',
                        help=('Print label with human readable'
                              'process name'))
    parser.add_argument('--copies', default=1, type=int,
                        help=('Number of printout copies, only used'
                              ' if neither container_name nor container_id'
                              ' type labels are printed. In that case, print'
                              ' one label of each type for each container.'))
    parser.add_argument('--pid',
                        help='The process LIMS id.')
    parser.add_argument('--log',
                        help='File name to use as log file')
    parser.add_argument('--use_printer',action="store_true",
                        help=('Print file on default or '
                              'supplied printer using lp command.'))
    parser.add_argument('--hostname',
                        help='Hostname for lp CUPS server.')
    parser.add_argument('--destination',
                        help='Name of printer.')
    parser.add_argument('--no_prepend', action='store_true',
                        help='Do not prepend old log, useful when ran locally')
    return parser.parse_args()


def main(args,lims,epp_logger):
    p = Process(lims,id=args.pid)
    lines = []
    cs = []
    if args.container_id:
        cs = p.output_containers()
        for c in cs:
            logging.info('Constructing barcode for container {0}.'.format(c.id))
            lines += makeContainerBarcode(c.id, copies=1)
    if args.container_name:
        cs = p.output_containers()
        for c in cs:
            logging.info('Constructing name label for container {0}.'.format(c.id))
            lines += makeContainerNameBarcode(c.name,copies=1)
    if args.operator_and_date:
        op = p.technician.name
        date = str(datetime.date.today())
        if cs: # list of containers
            copies = len(cs)
        else:
            copies = args.copies
        lines += makeOperatorAndDateBarcode(op,date,copies=copies)
    if args.process_name:
        pn = p.type.name
        if cs: # list of containers
            copies = len(cs)
        else:
            copies = args.copies
        lines += makeProcessNameBarcode(pn,copies=copies)
    if not (args.container_id or args.container_name or 
            args.operator_and_date or args.process_name):
        logging.info('No recognized label type given, exiting.')
        sys.exit(-1)
    if not args.use_printer:
        logging.info('Writing to stdout.')
        epp_logger.saved_stdout.write('\n'.join(lines)+'\n')
    elif lines: # Avoid printing empty files
        lp_args = ["lp"]
        if args.hostname:
            #remove that when all the calls to this script have been updated
            if args.hostname == 'homer.scilifelab.se:631':
                args.hostname='homer2.scilifelab.se:631'
            lp_args += ["-h",args.hostname]
        if args.destination:
            lp_args += ["-d",args.destination]
        lp_args.append("-") # lp accepts stdin if '-' is given as filename
        logging.info('Ready to call lp for printing.')
        sp = subprocess.Popen(lp_args, 
                              stdin=subprocess.PIPE, 
                              stdout=subprocess.PIPE, 
                              stderr=subprocess.PIPE)
        sp.stdin.write(str('\n'.join(lines)))
        logging.info('lp command is called for printing.')
        stdout,stderr = sp.communicate() # Will wait for sp to finish
        logging.info('lp stdout: {0}'.format(stdout))
        logging.info('lp stderr: {0}'.format(stderr))
        logging.info('lp command finished')
        sp.stdin.close()

if __name__ == '__main__':
    arguments = getArgs()
    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()
    prepend = not arguments.no_prepend
    with EppLogger(arguments.log,lims=lims,prepend=prepend) as epp_logger:
        main(arguments,lims,epp_logger)



#!/usr/bin/env python
DESC="""EPP script to fetch and upload Caliper image files for Clarity LIMS.
Searches the directory given by the path argument for filenames matching
a specific pattern ending with:
${INPUT.CONTAINER.PLACEMENT}_${INPUT.NAME}_${INPUT.CONTAINER.LIMSID}_${INPUT.LIMSID}.
This is done for each artifact of type ResultFile, that is of type PerInput. Any 
file found matching is copied to the current working directory with a name suffixed with the 
output artifact it is connected to. When executed as an EPP, this will cause the 
Clarity LIMS EPP wrapper to associate the file with this artifact.

Written by Johannes Alneberg, Science for Life Laboratory, Stockholm, Sweden.
"""

from genologics.lims import Lims
from genologics.entities import Artifact, Process,Container, Sample
from genologics.config import BASEURI,USERNAME,PASSWORD

from argparse import ArgumentParser
import os
import re
import sys
import logging

from genologics.epp import attach_file,EppLogger, unique_check, EmptyError

def main(lims, args, epp_logger):
    p = Process(lims, id=args.pid)

    if not args.path:
        args.path = os.getcwd()

    file_list = os.listdir(args.path)
    
    # Find all per input result files
    io = p.input_output_maps
    io_filtered = filter(lambda (x,y): y['output-generation-type']=='PerInput', io)
    io_filtered = filter(lambda (x,y): y['output-type']=='ResultFile', io_filtered)
    
    artifact_missing_file = []
    artifact_multiple_file = []
    found_files = []

    for input, output in io_filtered:
        i_a = Artifact(lims,id=input['limsid'])
        o_a = Artifact(lims,id=output['limsid'])

        # Input Well, Input Container
        i_w, i_c = i_a.location[1], i_a.location[0]

        # Well is typed without colon in filename:
        i_w = ''.join(i_w.split(':'))
        
        info = {'well':i_w,
                'container_id':i_c.id,
                'input_artifact_id':i_a.id}

        # Use a reguluar expression to find the file name given
        # the container and sample. This is all assuming the driver template name ends with:
        # ${INPUT.CONTAINER.PLACEMENT}_${INPUT.NAME}_${INPUT.CONTAINER.LIMSID}_${INPUT.LIMSID}
        # However, names are excluded to improve robustness.
        re_str = '.*{well}_.*_.*{container_id}_.*{input_artifact_id}'\
                                   .format(**info)

        im_file_r = re.compile(re_str)
        fns = filter(im_file_r.match, file_list)
        logging.info(("Looking for file for artifact id: {input_artifact_id} "
                      "from container with id: {container_id}.").format(**info))

        if len(fns) == 0:
            logging.warning("No image file found for artifact with id {0}".format(i_a.id))
            artifact_missing_file.append(i_a)
        elif len(fns) > 1:
            logging.warning(("Multiple image files found for artifact with id {0}, "
                            "please attach files manually").format(i_a.id))
            artifact_multiple_file.append(i_a)
        else:
            fn = fns[0]
            found_files.append(fn)
            logging.info("Found image file {0} for artifact with id {1}".format(fn, i_a.id))
            fp = os.path.join(args.path, fn)
            
            # Attach file to the LIMS
            location = attach_file(fp, o_a)
            logging.debug("Moving {0} to {1}".format(fp,location))

    warning = ""
    if len(artifact_missing_file):
        warning = "Did not find any file for {0} artifact(s). ".format(len(artifact_missing_file))

    if len(artifact_multiple_file):
        warning += "Found multiple files for {0} artifact(s), none of these were uploaded.".format(len(artifact_multiple_file))
    
    if warning:
       warning = "Warning: " + warning

    abstract = "Uploaded {0} file(s). {1}".format(len(found_files), warning)
    print >> sys.stderr, abstract # stderr will be logged and printed in GUI


if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log',
                        help='Log file for runtime info and errors')
    parser.add_argument('--path',
                        help='Path where image files are located')
    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()

    with EppLogger(args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)




#!/usr/bin/env python
DESC="""EPP script to aggregate the number of reads from different demultiplexing runs,
based on the flag 'include reads' located at the same level as '# reads' 

Denis Moreno, Science for Life Laboratory, Stockholm, Sweden
""" 
from argparse import ArgumentParser
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.epp import attach_file, EppLogger
import logging
import sys
import os
from genologics.entities import *


DEMULTIPLEX={'13' : 'Bcl Conversion & Demultiplexing (Illumina SBS) 4.0'}
SUMMARY = {'356' : 'Project Summary 1.3'}
SEQUENCING = {'38' : 'Illumina Sequencing (Illumina SBS) 4.0','46' : 'MiSeq Run (MiSeq) 4.0'}
def main(lims, args, logger):
    """This should be run at project summary level"""
    p = Process(lims,id = args.pid)
    samplenb=0
    errnb=0
    summary={}
    logart=None
    for output_artifact in p.all_outputs():
        #filter to only keep solo sample demultiplexing output artifacts
        if output_artifact.type=='Analyte' and len(output_artifact.samples)==1:
            sample=output_artifact.samples[0]
            samplenb+=1
            #update the total number of reads
            total_reads=sumreads(sample, summary)
            sample.udf['Total Reads (M)']=total_reads
            output_artifact.udf['Set Total Reads']=total_reads
            logging.info("Total reads is {0} for sample {1}".format(sample.udf['Total Reads (M)'],sample.name))
            try:
                if sample.udf['Reads Min'] >= sample.udf['Total Reads (M)']:
                    sample.udf['Status (auto)']="In Progress"
                    sample.udf['Passed Sequencing QC']="False"
                elif sample.udf['Reads Min'] < sample.udf['Total Reads (M)'] : 
                    sample.udf['Passed Sequencing QC']="True"
                    sample.udf['Status (auto)']="Finished"
            except KeyError as e:
                print e
                logging.warning("No reads minimum found, cannot set the status auto flag for sample {0}".format(sample.name))
                errnb+=1

            #commit the changes
            sample.put()
            output_artifact.put()
        elif(output_artifact.type=='Analyte') and len(output_artifact.samples)!=1:
            logging.error("Found {0} samples for the ouput analyte {1}, that should not happen".format(len(output_artifact.samples()),output_artifact.id))
        elif(output_artifact.type=="ResultFile" and output_artifact.name=="AggregationLog"):
            logart=output_artifact


    #write the csv file, separated by pipes, no cell delimiter
    with open("AggregationLog.csv", "w") as f:
        f.write("sep=,\n")
        f.write('sample name,number of flowcells,number of lanes,flowcell1:lane1|lane2;flowcell2:lane1|lane2|lane3 ...\n')
        for sample in summary:
            view=[]
            totfc=len(summary[sample])
            totlanes=0
            for fc in summary[sample]:
                view.append("{0}:{1}".format(fc, "|".join(summary[sample][fc])))
                totlanes+=len(summary[sample][fc])
            f.write('{0},{1},{2},{3}\n'.format(sample, totfc, totlanes, ";".join(view)))
    try:
        attach_file(os.path.join(os.getcwd(), "AggregationLog.csv"), logart)
        logging.info("updated {0} samples with {1} errors".format(samplenb, errnb))
    except AttributeError:
        #happens if the log artifact does not exist, if the step has been started before the configuration changes
        logging.info("Could not upload the log file")
            
def demnumber(sample):
    """Returns the number of distinct demultiplexing processes tagged with "Include reads" for a given sample"""
    expectedName="{0} (FASTQ reads)".format(sample.name)
    dem=set()
    arts=lims.get_artifacts(sample_name=sample.name,process_type=DEMULTIPLEX.values(), name=expectedName)   
    for a in arts:
        if a.udf["Include reads"] == "YES":
            dem.add(a.parent_process.id)
    return len(dem)
    
def sumreads(sample, summary):
    if sample.name not in summary:
        summary[sample.name]={}
    expectedName="{0} (FASTQ reads)".format(sample.name)
    arts=lims.get_artifacts(sample_name=sample.name,process_type=DEMULTIPLEX.values(), name=expectedName)   
    tot=0
    fclanel=[]
    filteredarts=[]
    base_art=None
    for a in sorted(arts, key=lambda art:art.parent_process.date_run, reverse=True):
        if "# Reads" not in a.udf:
            continue
        try:
            if 'Include reads' in a.udf:
                orig=getParentInputs(a)
                for o in orig:
                    if sample in o.samples:
                        fc="{0}:{1}".format(o.location[0].name,o.location[1].split(":")[0])
                        if fc not in fclanel:
                            filteredarts.append(a)
                            fclanel.append(fc)
                        if o.location[0].name in summary[sample.name]:
                            summary[sample.name][o.location[0].name].add(o.location[1].split(":")[0])
                        else:
                            summary[sample.name][o.location[0].name]=set(o.location[1].split(":")[0])

        except KeyError:
            #Happens if the "Include reads" does not exist
            pass

    for i in xrange(0,len(filteredarts)):    
        a=filteredarts[i]
        if a.udf['Include reads']=='YES':
            base_art=a
            tot+=float(a.udf['# Reads'])

    #grab the sequencing process associated 
    #find the correct input
    inputart=None
    try:
        for inart in base_art.parent_process.all_inputs():
            if sample.name in [s.name for s in inart.samples]:
                try:
                    sq=lims.get_processes(type=SEQUENCING.values(), inputartifactlimsid=inart.id)[0]
                except TypeError:
                    logging.error("Did not manage to get sequencing process for artifact {0}".format(inart.id))
                else:
                    if "Read 2 Cycles" in sq.udf and sq.udf['Read 2 Cycles'] is not None:
                        tot/=2
                break
    except AttributeError as e:
        print e
        #base_art is still None because no arts were found
        logging.info("No demultiplexing processes found for sample {0}".format(sample.name))



    # total is displayed as millions
    tot/=1000000
    return tot

def getParentInputs(art):
    inp=set()
    for i in art.parent_process.input_output_maps:
        if i[1]['uri'].id==art.id:
            inp.add(i[0]['uri'])

    return inp
    

if __name__=="__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log',
                        help='Log file for runtime info and errors.')
    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()

    with EppLogger(args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)




#!/usr/bin/env python
import os
import subprocess
from argparse import ArgumentParser

TEMPLATE = """
Scripts in the Genologics Package
=================================
Short usage descriptions for the scripts in the Genologics Package.

"""

def indent(s):
    return "\n".join(map(lambda w: '\t'+w, s.splitlines()))


def help_doc_rst(script,file_path):
    """ Fetch the --help info from the script, outputs rst formatted string """
    script_path = os.path.join(file_path,script)
    sp = subprocess.Popen(["python",script_path,"--help"],
                          stdout=subprocess.PIPE)
    stdout,stderr = sp.communicate()
    
    # Add help message to template
    header = ("{0}\n{1}\nAutomated help message generated from running {0} "
              "with the --help flag::\n\n").format(script,'-'*len(script))

    return header + "{0}\n\n".format(indent(stdout))
    

if __name__ == "__main__":
    # Argumentparser only to add --help option
    args = ArgumentParser(description=("Generates basic documentation on all scripts "
                                       "contained in the scripts folder. Used instead "
                                       "of sphinx extension since readthedocs build "
                                       "failed on genologics imports. ")).parse_args()

    file_path = os.path.dirname(os.path.realpath(__file__))
    this_file = os.path.basename(__file__)

    # Fetch all python scripts in the scripts folder
    file_list = os.listdir(file_path)
    scripts = filter(lambda fn: fn[-3:]=='.py' and fn != this_file, file_list)
    scripts = sorted(scripts)

    for script in scripts:
        TEMPLATE += help_doc_rst(script, file_path)

    # This script added last:
    TEMPLATE += help_doc_rst(this_file, file_path)

    # Print all help messages to a sphinx (restructured text) markup file
    docs_path = os.path.join(file_path,'..','docs','scripts.rst')
    with open(docs_path,'w') as doc_f:
        doc_f.write(TEMPLATE)
    



#!/usr/bin/env python
DESC = """EPP script for Quant-iT measurements to calculate concentrations and 
load input artifact-udfs and output file-udfs of the process with concentration 
values and fluorescence intensity.

Reads from:
    --Lims fields--
    "Assay type"                process udf 
    "Standard volume"           process udf
    "Standard dilution"         process udf
    "Linearity of standards"    process udf
    "Sample volume"             process udf

    --files--
    "Standards File (.txt)"     "shared result file" uploaded by user.   
    "Quant-iT Result File 1"    "shared result file" uploaded by user.
    "Quant-iT Result File 2"    "shared result file" uploaded by user. (optional)

Writes to:
    --Lims fields--
    "Fluorescence intensity 1"  udf of input analytes to the process
    "Fluorescence intensity 2"  udf of input analytes to the  process
    "Concentration"             udf of process artifacts (result file)
    "Conc. Units"               udf of process artifacts (result file)

Logging:
    The script outputs a regular log file with regular execution information.

Performance:
    1)  Makes a linear regression analysis on the End RFU standards read from 
    "Standards File (.txt)" --> slope, intersect and R2 (Pearson correlation 
    coefficient)

    2)  If R2 is less than "Linearity of standards", standards are not ok and 
    the user is warned "Problem with standards! Redo measurement!". 

    3)  If R2 is greater than "Linearity of standards", concentration values are
    calculated based on the linear regression parameters, copied to the 
    "Concentration"-udf on the output artifacts (one result file per sample). 
    The "Conc. Units"-udf on the same artifacts are set to "ng/ul".

    4)  "End RFU" values from "Quant-iT Result File 1" are copied to the udf
    "Fluorescence intensity 1" of the input analyte to the process, and 
    "End RFU" values from "Quant-iT Result File 2" are copied to the udf 
    "Fluorescence intensity 2". 

Written by Maya Brandi 
"""

import os
import sys
import logging
import numpy as np

from argparse import ArgumentParser
from requests import HTTPError
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import set_field
from genologics.epp import ReadResultFiles

class QuantitConc():
    def __init__(self, process, file_handler):
        self.file_handler = file_handler
        self.udfs = dict(process.udf.items())
        self.abstract = []
        self.missing_udfs = []
        self.missing_samps =[]
        self.no_samps = 0
        self.standards = {}
        self.model = []
        self.result_files = {}

    def prepare_result_files_dict(self):
        """"Quant-iT Result File 1" and "Quant-iT Result File 2" (the second is 
        optional) uploaded by user, are formated and stored in a dict"""
        result_files_dict = {}
        file_names = {'Quant-iT Result File 1':'Fluorescence intensity 1',
                    'Quant-iT Result File 2':'Fluorescence intensity 2'}
        for f_name, udf_name in file_names.items():
            if self.file_handler.shared_files.has_key(f_name):
                result_file = self.file_handler.shared_files[f_name]
                result_files_dict[udf_name] = self.file_handler.format_file(
                    result_file, name = f_name, root_key_col = 1, header_row = 19)
        self.result_files = result_files_dict

    def _make_standards_dict(self):
        """End RFU standards are read from 'Standards File (.txt)' and stored 
        in a dict"""
        standards_file = self.file_handler.shared_files['Standards File (.txt)']
        standards_file_formated = self.file_handler.format_file(standards_file, 
                name = 'Standards File (.txt)', root_key_col = 1, header_row = 19)
        standards_dict = {}
        for k,v in standards_file_formated.items():
            cond1 = set(['Sample','End RFU']).issubset(v) 
            cond2 = v['Sample'].split()[0]=='Standard'
            if cond1 and cond2:
                standard = int(v['Sample'].split()[1])
                standards_dict[standard] = float(v['End RFU'])
        return standards_dict

    def _nuclear_acid_amount_in_standards(self):
        """Nuclear acid amount in standards is calculated as
        "Supposed concentrations of standards" * "Standard volume" / "Standard 
        dilution" Supposed concentrations of standards are different between 
        Quant-iT BR assays and Quant-iT HS assays."""
        requiered_udfs = set(['Standard volume','Assay type','Standard dilution'])
        supp_conc_stds = {'RNA BR':[0,5,10,20,40,60,80,100],
                          'DNA BR':[0,5,10,20,40,60,80,100],
                          'RNA':[0,0.5,1,2,4,6,8,10],
                          'DNA HS':[0,0.5,1,2,4,6,8,10]}    
        if requiered_udfs.issubset(self.udfs.keys()):
            nuclear_acid_amount = np.ones(8)
            for standard in range(8):
                supp_conc_standard = supp_conc_stds[self.udfs['Assay type']][standard]
                stand_conc = supp_conc_standard * self.udfs['Standard volume']
                stand_dil = self.udfs['Standard dilution']
                nuclear_acid_amount[standard] = np.true_divide(stand_conc, stand_dil)
            return nuclear_acid_amount
        elif not requiered_udfs.issubset(self.missing_udfs):
            self.missing_udfs += requiered_udfs
        return None

    def _linear_regression(self, X,Y):
        """Perform linear regression with intersect forced to origin. And 
        calculates Pearson correlation coefficient R2. """
        X = np.array(X)
        X_force_zero = X[:,np.newaxis]
        slope = np.linalg.lstsq(X_force_zero, Y)[0]

        A = np.array([ X, np.ones(len(X))])
        resid = np.linalg.lstsq(A.T,Y)[1]
        R2 = 1 - resid / (Y.size * Y.var())
        return float(R2), float(slope)

    def fit_model(self):
        """Performing linear regression on standards.
        X = Relative fluorescence intensities of 8 standards 
        Y = Nuclear acid amount in standards 
        Y = slope*X + intersect; R2=Pearson correlation coefficient."""
        self.standards = self._make_standards_dict()
        amount_in_standards = self._nuclear_acid_amount_in_standards()
        if amount_in_standards is not None:
            relative_standards = np.ones(8)
            for k,v in self.standards.items(): 
                relative_standards[k-1] = v - self.standards[1]
            R2, slope = self._linear_regression(relative_standards, 
                            amount_in_standards)
            self.model = [R2, slope]
        else:
            self.model = None

    def get_and_set_fluor_int(self, target_file):
        """Copies "End RFU" values from "Quant-iT Result File 1" and 
        "Quant-iT Result File 2" (if provided) to udfs "Fluorescence intensity 1"
        and "Fluorescence intensity 2. Calculates and returns Relative 
        fluorescence intensity standards:
        rel_fluor_int = The End RFU of standards - Background fluorescence 
        intensity"""
        sample = target_file.samples[0].name
        fluor_int = []
        target_udfs = target_file.udf
        # For the moment we dont know ofa way to delete udfs. Should be solved.
        #if dict(target_udfs.items()).has_key('Fluorescence intensity 1'):   
        #    del target_udfs['Fluorescence intensity 1']                   
        #if dict(target_udfs.items()).has_key('Fluorescence intensity 2'): 
        #    del target_udfs['Fluorescence intensity 2']
        target_file.udf = target_udfs
        for udf_name ,formated_file in self.result_files.items():
            if sample in formated_file.keys():
                fluor_int.append(float(formated_file[sample]['End RFU']))
                target_file.udf[udf_name] = float(formated_file[sample]['End RFU']) 
            else:
                self.missing_samps.append(sample)
        set_field(target_file)
        rel_fluor_int = np.mean(fluor_int) - self.standards[1]
        return rel_fluor_int

    def calc_and_set_conc(self, target_file, rel_fluor_int):
        """Concentrations are calculated based on the linear regression 
        parametersand copied to the "Concentration"-udf of the target_file. The 
        "Conc. Units"-udf is set to "ng/ul"""
        requiered_udfs = set(['Sample volume','Standard dilution','WS volume'])
        if requiered_udfs.issubset(self.udfs.keys()) and self.model:
            conc = np.true_divide((self.model[1] * rel_fluor_int * (
                self.udfs['WS volume'] + self.udfs['Sample volume'])),
                self.udfs['Sample volume']*(self.udfs['WS volume'] + 
                self.udfs['Standard volume']))
            target_file.udf['Concentration'] = conc
            target_file.udf['Conc. Units'] = 'ng/ul'
            set_field(target_file)
            self.no_samps +=1
        elif not requiered_udfs.issubset(self.missing_udfs):
            self.missing_udfs += requiered_udfs

    
def main(lims, pid, epp_logger):
    process = Process(lims,id = pid)
    target_files = dict((r.samples[0].name, r) for r in process.result_files())
    file_handler = ReadResultFiles(process)
    QiT = QuantitConc(process, file_handler)
    QiT.fit_model()
    QiT.prepare_result_files_dict()
    if QiT.model and 'Linearity of standards' in QiT.udfs.keys():
        R2 = QiT.model[0]
        if R2 >= QiT.udfs['Linearity of standards']:
            QiT.abstract.insert(0,"R2 = {0}. Standards OK.".format(R2))
            if QiT.result_files:
                for sample, target_file in target_files.items():
                    rel_fluor_int = QiT.get_and_set_fluor_int(target_file)
                    QiT.calc_and_set_conc(target_file, rel_fluor_int)
                QiT.abstract.append("Concentrations uploaded for {0} "
                                                "samples.".format(QiT.no_samps))
            else:
                QiT.abstract.append("Upload input file(s) for samples.")
        else:
            QiT.abstract.insert(0, "R2 = {0}. Problem with standards! Redo "
                                                    "measurement!".format(R2))
    else:
        QiT.missing_udfs.append('Linearity of standards')
    if QiT.missing_samps:
        QiT.abstract.append("The following samples are missing in Quant-iT "
            "result File 1 or 2: {0}.".format(', '.join(QiT.missing_samps)))
    if QiT.missing_udfs:
        QiT.abstract.append("Are all of the following udfs set? : {0}".format(
                                                   ', '.join(QiT.missing_udfs)))
    print >> sys.stderr, ' '.join(QiT.abstract)

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = None , dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))

    args = parser.parse_args()
    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, epp_logger)



#!/usr/bin/env python
DESC = """EPP script to calculate molar concentration given the 
weight concentration, in Clarity LIMS. Before updating the artifacts, 
the script verifies that 'Concentration' and 'Size (bp)' udf:s are not blank,
 and that the 'Conc. units' field is 'ng/ul' for each artifact. Artifacts 
that do not fulfill the requirements, will not be updated.

Written by Johannes Alneberg, Science for Life Laboratory, Stockholm, Sweden
""" 

from argparse import ArgumentParser

from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD

from genologics.entities import Process
from genologics.epp import EppLogger

import logging
import sys

def apply_calculations(lims, artifacts, conc_udf, size_udf, unit_udf, epp_logger):
    for artifact in artifacts:
        logging.info(("Updating: Artifact id: {0}, "
                     "Concentration: {1}, Size: {2}, ").format(artifact.id, 
                                                        artifact.udf[conc_udf],
                                                        artifact.udf[size_udf]))
        factor = 1e6 / (328.3 * 2 * artifact.udf[size_udf])
        artifact.udf[conc_udf] = artifact.udf[conc_udf] * factor
        artifact.udf[unit_udf] = 'nM'
        artifact.put()
        logging.info('Updated {0} to {1}.'.format(conc_udf,
                                                 artifact.udf[conc_udf]))
def check_udf_is_defined(artifacts, udf):
    """ Filter and Warn if udf is not defined for any of artifacts. """
    filtered_artifacts = []
    incorrect_artifacts = []
    for artifact in artifacts:
        if (udf in artifact.udf):
            filtered_artifacts.append(artifact)
        else:
            logging.warning(("Found artifact for sample {0} with {1} "
                             "undefined/blank, skipping").format(artifact.samples[0].name, udf))
            incorrect_artifacts.append(artifact)
    return filtered_artifacts, incorrect_artifacts


def check_udf_has_value(artifacts, udf, value):
    """ Filter artifacts on undefined udf or if udf has wrong value. """
    filtered_artifacts = []
    incorrect_artifacts = []
    for artifact in artifacts:
        if udf in artifact.udf and (artifact.udf[udf] == value):
            filtered_artifacts.append(artifact)
        elif udf in artifact.udf:
            incorrect_artifacts.append(artifact)
            logging.warning(("Filtered out artifact for sample: {0}"
                          ", due to wrong {1}").format(artifact.samples[0].name, udf))
        else:
            incorrect_artifacts.append(artifact)
            logging.warning(("Filtered out artifact for sample: {0}"
                          ", due to undefined/blank {1}").format(artifact.samples[0].name, udf))

    return filtered_artifacts, incorrect_artifacts

def main(lims, args, epp_logger):
    p = Process(lims, id = args.pid)
    udf_check = 'Conc. Units'
    value_check = 'ng/ul'
    concentration_udf = 'Concentration'
    size_udf = 'Size (bp)'

    if args.aggregate:
        artifacts = p.all_inputs(unique=True)
    else:
        all_artifacts = p.all_outputs(unique=True)
        artifacts = filter(lambda a: a.output_type == "ResultFile", all_artifacts)

    correct_artifacts, no_concentration = check_udf_is_defined(artifacts, concentration_udf)
    correct_artifacts, no_size = check_udf_is_defined(correct_artifacts, size_udf)
    correct_artifacts, wrong_value = check_udf_has_value(correct_artifacts, udf_check, value_check)

    apply_calculations(lims, correct_artifacts, concentration_udf, size_udf, udf_check, epp_logger)

    d = {'ca': len(correct_artifacts),
         'ia': len(wrong_value) + len(no_size) + len(no_concentration)}

    abstract = ("Updated {ca} artifact(s), skipped {ia} artifact(s) with "
                "wrong and/or blank values for some udfs.").format(**d)

    print >> sys.stderr, abstract # stderr will be logged and printed in GUI


if __name__ == "__main__":
    # Initialize parser with standard arguments and description
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', default=sys.stdout,
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))
    parser.add_argument('--aggregate', action='store_true',
                        help=("Use this tag if your process is aggregating "
                              "results. The default behaviour assumes it is "
                              "the output artifact of type analyte that is "
                              "modified while this tag changes this to using "
                              "input artifacts instead"))

    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()
    with EppLogger(args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)



#!/usr/bin/env python
DESC = """EPP script to copy user defined fields from analyte level to 
submitted sample level in Clarity LIMS. Can be executed in the background 
or triggered by a user pressing a "blue button".

This script can only be applied to processes where ANALYTES are modified 
in the GUI. The script can output two different logs, where the 
status_changelog contains notes with the technician, the date and changed 
status for each copied status. The regular log file contains regular 
execution information. 

Error handling:
If the udf given is blank or not defined for any of the inputs,
the script will log this, and not perform any changes for that artifact.


Written by Johannes Alneberg, Science for Life Laboratory, Stockholm, Sweden
""" 
import os
import sys
import logging

from argparse import ArgumentParser

from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import CopyField


def main(lims, args, epp_logger):
    source_udfs = args.source_udf
    dest_udfs = args.dest_udf
    correct_artifacts = 0
    incorrect_artifacts = 0
    no_updated = 0
    p = Process(lims,id = args.pid)
    artifacts, inf = p.analytes()


    if args.status_changelog:
        epp_logger.prepend_old_log(args.status_changelog)

    if not dest_udfs:
        dest_udfs = source_udfs
    elif len(dest_udfs) != len(source_udfs):
        logging.error("source_udfs and dest_udfs lists of arguments are uneven.")
        sys.exit(-1)
    for i in range(len(source_udfs)):
        source_udf = source_udfs[i]
        dest_udf = dest_udfs[i]
        with open(args.status_changelog, 'a') as changelog_f:
            for artifact in artifacts:
                if source_udf in artifact.udf:
                    correct_artifacts = correct_artifacts +1
                    copy_sesion = CopyField(artifact, artifact.samples[0], source_udf, dest_udf)
                    test = copy_sesion.copy_udf(changelog_f)
                    if test:
                        no_updated = no_updated + 1
                else:
                    incorrect_artifacts = incorrect_artifacts + 1
                    logging.warning(("Found artifact for sample {0} with {1} "
                                   "undefined/blank, exiting").format(artifact.samples[0].name, source_udf))

    if incorrect_artifacts == 0:
        warning = "no artifacts"
    else:
        warning = "WARNING: skipped {0} udfs(s)".format(incorrect_artifacts)
    d = {'ua': no_updated,
         'ca': correct_artifacts,
         'ia': incorrect_artifacts,
         'warning' : warning}

    abstract = ("Updated {ua} udf(s), out of {ca} in total, "
                "{warning} with incorrect udf info.").format(**d)

    print >> sys.stderr, abstract # stderr will be logged and printed in GUI


if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log',
                        help=('File name for standard log file, '
                              ' for runtime information and problems.'))
    parser.add_argument('-s', '--source_udf', type=str, default=None, nargs='*',
                        help=('Name(s) of the source user defined field(s) '
                               'that will be copied. One or many udf-names '
                               'can be given.'))
    parser.add_argument('-d', '--dest_udf', type=str, default=None, nargs='*',
                        help=('Name(s) of the destination user defined '
                              'field(s) that will be written to. This '
                              'argument is optional, if left empty '
                              'the source_udf argument is used instead. '
                              'Zero or many udf-names can be given. If '
                              'more than zero, the numer of udfs needs '
                              'to be the same as number of source_udfs'))
    parser.add_argument('-c', '--status_changelog',
                        help=('File name for status changelog file, '
                              ' for concise information on who, what and '
                              ' when for status change events. '
                              'Prepends the old changelog file by default.'))
    args = parser.parse_args()

    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)



"""Python interface to GenoLogics LIMS via its REST API.

Entities and their descriptors for the LIMS interface.

Per Kraulis, Science for Life Laboratory, Stockholm, Sweden.
Copyright (C) 2012 Per Kraulis
"""

import re
import urlparse
import datetime
import time
from xml.etree import ElementTree
import logging

logger = logging.getLogger(__name__)

_NSMAP = dict(
    art='http://genologics.com/ri/artifact',
    artgr='http://genologics.com/ri/artifactgroup',
    cnf='http://genologics.com/ri/configuration',
    con='http://genologics.com/ri/container',
    ctp='http://genologics.com/ri/containertype',
    exc='http://genologics.com/ri/exception',
    file='http://genologics.com/ri/file',
    inst='http://genologics.com/ri/instrument',
    lab='http://genologics.com/ri/lab',
    prc='http://genologics.com/ri/process',
    prj='http://genologics.com/ri/project',
    prop='http://genologics.com/ri/property',
    protcnf='http://genologics.com/ri/protocolconfiguration',
    protstepcnf='http://genologics.com/ri/stepconfiguration',
    prx='http://genologics.com/ri/processexecution',
    ptm='http://genologics.com/ri/processtemplate',
    ptp='http://genologics.com/ri/processtype',
    res='http://genologics.com/ri/researcher',
    ri='http://genologics.com/ri',
    rt='http://genologics.com/ri/routing',
    rtp='http://genologics.com/ri/reagenttype',
    smp='http://genologics.com/ri/sample',
    stg='http://genologics.com/ri/stage',
    stp='http://genologics.com/ri/step',
    udf='http://genologics.com/ri/userdefined',
    ver='http://genologics.com/ri/version',
    wkfcnf='http://genologics.com/ri/workflowconfiguration')

for prefix, uri in _NSMAP.iteritems():
    ElementTree._namespace_map[uri] = prefix

_NSPATTERN = re.compile(r'(\{)(.+?)(\})')

def nsmap(tag):
    "Convert from normal XML-ish namespace tag to ElementTree variant."
    parts = tag.split(':')
    if len(parts) != 2:
        raise ValueError("no namespace specifier in tag")
    return "{%s}%s" % (_NSMAP[parts[0]], parts[1])


class SampleHistory:
    """Class handling the history generation for a given sample/artifact
    AFAIK the only fields of the history that are read are proc.type and outart""" 

    def __init__(self, sample_name=None, output_artifact=None, input_artifact=None, lims=None, pro_per_art=None, test=False):
        self.processes_per_artifact=pro_per_art
        if lims:
            self.lims = lims
            if not (test):
                #this is now the default
                self.sample_name=sample_name
                self.alternate_history(output_artifact, input_artifact)
                self.art_map=None
            elif (sample_name) and pro_per_art:
                self.sample_name=sample_name
                self.make_sample_artifact_map()
                if output_artifact:
                    self.get_analyte_hist_sorted(output_artifact, input_artifact)
        else:
            logger.error("Tried to build History without lims")
            raise AttributeError("History cannot be computed without a valid lims object")


    def control(self):
        """this can be used to check the content of the object.
        """
        logger.info("SAMPLE NAME: {}".format(self.sample_name))
        logger.info("outart : {}".format(self.history_list[0]))
        #logger.info ("\nmap :")
        #for key, value in self.art_map.iteritems():
        #    logger.info(value[1]+"->"+value[0].id+"->"+key)
        logger.info ("\nHistory :\n\n")
        logger.info("Input\tProcess\tProcess info")
        for key, dict in self.history.iteritems():
            logger.info (key)
            for key2, dict2 in dict.iteritems():
                logger.info ("\t{}".format(key2))
                for key, value in dict2.iteritems():
                    logger.info ("\t\t{0}->{1}".format(key,(value if value is not None else "None")))
        logger.info ("\nHistory List")
        for art in self.history_list:
            logger.info (art)
        
    def make_sample_artifact_map(self):
        """samp_art_map: connects each output artifact for a specific sample to its 
        corresponding process and input artifact assuming, for a given sample,
        one input -> one process -> one output
        This function starts from the output, 
        and creates an entry like this : output -> (process, input)"""
        samp_art_map ={}
        if self.sample_name:
            artifacts = self.lims.get_artifacts(sample_name = self.sample_name, type = 'Analyte', resolve=False) 
            for one_art in artifacts:
                input_arts = one_art.input_artifact_list()
                for input_art in input_arts:
                    for samp in input_art.samples:
                        if samp.name == self.sample_name:
                            samp_art_map[one_art.id] = (one_art.parent_process, input_art.id)

        self.art_map=samp_art_map
    def alternate_history(self, out_art, in_art=None):
        """This is a try at another way to generate the history.
        This one iterates over Artifact.parent_process and Process.all_inputs()
        Then, it takes all the child processes for each input (because we want
        qc processes too) and puts everything in a dictionnary. 
        """
        history = {}
        hist_list = []
       #getting the list of all expected analytes.
        artifacts = self.lims.get_artifacts(sample_name = self.sample_name, type = 'Analyte', resolve=False)
        processes=[]
        inputs=[]
        if in_art:
            #If theres an input artifact given, I need to make a loop for this one, before treating it as an output
            starting_art=in_art
            inputs.append(in_art)
            history[in_art]={}
            #If there is a loacl map, use it. else, query the lims.
            if self.processes_per_artifact and in_art in self.processes_per_artifact:
                valid_pcs=self.processes_per_artifact[in_art]
            else:
                valid_pcs=self.lims.get_processes(inputartifactlimsid=in_art)

            for tempProcess in valid_pcs:
                history[in_art][tempProcess.id] = {'date' : tempProcess.date_run,
                                                   'id' : tempProcess.id,
                                                   'outart' : (out_art if out_art in [ out.id for out in tempProcess.all_outputs()] else None ),
                                                   'inart' : in_art,
                                                   'type' : tempProcess.type.id,
                                                   'name' : tempProcess.type.name}
        else:
            starting_art=out_art
        #main iteration    
        #it is quite heavy on logger at info level
        not_done=True
        while not_done:
            logger.info ("looking for "+(starting_art))
            not_done=False 
            for o in artifacts:
                logger.info (o.id)
                if o.id == starting_art:
                    if o.parent_process is None:
                        #flow control : if there is no parent process, we can stop iterating, we're done.
                        not_done=False
                        break #breaks the for artifacts, we are done anyway.
                    else:
                        not_done=True #keep the loop running
                    logger.info ("found it")
                    processes.append(o.parent_process)
                    logger.info ("looking for inputs of "+o.parent_process.id)
                    for i in o.parent_process.all_inputs():
                        logger.info (i.id)
                        if i in artifacts:
                            history[i.id]={}
                            for tempProcess in (self.processes_per_artifact[i.id] if self.processes_per_artifact else self.lims.get_processes(inputartifactlimsid=i.id)):#If there is a loacl map, use it. else, query the lims.
                                history[i.id][tempProcess.id] = {'date' : tempProcess.date_run,
                                                               'id' : tempProcess.id,
                                                               'outart' : (o.id if tempProcess.id == o.parent_process.id else None),
                                                               'inart' : i.id,
                                                               'type' : tempProcess.type.id,
                                                               'name' : tempProcess.type.name}



                            logger.info ("found input "+i.id)
                            inputs.append(i.id) #this will be the sorted list of artifacts used to rebuild the history in order
                            # while increment
                            starting_art=i.id
                            
                            break #break the for allinputs, if we found the right one
                    break # breaks the for artifacts if we matched the current one
        self.history=history
        self.history_list=inputs 


    def get_analyte_hist_sorted(self, out_artifact, input_art = None):
         """Makes a history map of an artifac, using the samp_art_map 
         of the corresponding sample.
         The samp_art_map object is built up from analytes. This means that it will not 
         contain output-input info for processes wich have only files as output. 
         This is logical since the samp_art_map object is used for building up the ANALYTE 
         history of a sample. If you want to make the analyte history based on a 
         resultfile, that is; if you want to give a resultfile as out_artifact here, 
         and be given the historylist of analytes and processes for that file, you 
         will also have to give the input artifact for the process that generated 
         the resultfile for wich you want to get the history. In other words, if you 
         want to get the History of the folowing scenario:        
 
         History --- > Input_analyte -> Process -> Output_result_file
         
         then the arguments to this function should be:
         out_artifact = Output_result_file
         input_art = Input_analyte
 
         If you instead want the History of the folowing scenario:
         
         History --- > Input_analyte -> Process -> Output_analyte
 
         then you can skip the input_art argument and only set:
         out_artifact = Output_analyte 
         """
         history = {}
         hist_list = []
         if input_art:
            # In_art = Artifact(lims,id=input_art)
            # try:
            #     pro = In_art.parent_process.id
            # except:
            #     pro = None
             history, out_artifact = self._add_out_art_process_conection_list(input_art, 
                                                         out_artifact, history)
             hist_list.append(input_art)
         while self.art_map.has_key(out_artifact):
             pro, input_art = self.art_map[out_artifact]
             hist_list.append(input_art)
             history, out_artifact = self._add_out_art_process_conection_list(input_art, 
                                                        out_artifact, history)
         self.history=history
         self.history_list=hist_list
 
    def _add_out_art_process_conection_list(self, input_art, out_artifact, history = {}):
        """This function populates the history dict with process info per artifact.
        Maps an artifact to all the processes where its used as input and adds this 
        info to the history dict. Observe that the output artifact for the input 
        artifact in the historychain is given as input to this function. All 
        processes that the input artifact has been involved in, but that are not 
        part of the historychain get the outart set to None. This is very important."""
        # Use the local process map if we have one, else, query the lims 
        for process in self.processes_per_artifact[input_art] if self.processes_per_artifact else lims.get_processes(inputartifactlimsid = inart):
            #outputs = map(lambda a: (a.id), process.all_outputs())
            outputs = [a.id for a in process.all_outputs()] 
            outart = out_artifact if out_artifact in outputs else None 
            step_info = {'date' : process.date_run,
                         'id' : process.id,
                         'outart' : outart,
                         'inart' : input_art,
                         'type' : process.type.id,
                         'name' : process.type.name}
            if history.has_key(input_art):
                history[input_art][process.id] = step_info
            else:
                history[input_art] = {process.id : step_info}
        return history, input_art

class BaseDescriptor(object):
    "Abstract base descriptor for an instance attribute."

    def __get__(self, instance, cls):
        raise NotImplementedError


class TagDescriptor(BaseDescriptor):
    """Abstract base descriptor for an instance attribute
    represented by an XML element.
    """

    def __init__(self, tag):
        self.tag = tag


class StringDescriptor(TagDescriptor):
    """An instance attribute containing a string value
    represented by an XML element.
    """

    def __get__(self, instance, cls):
        instance.get()
        node = self.get_node(instance)
        if node is None:
            return None
        else:
            return node.text

    def __set__(self, instance, value):
        node = self.get_node(instance)
        if node is None:
            raise AttributeError("no element '%s' to set" % self.tag)
        else:
            node.text = value

    def get_node(self, instance):
        if self.tag:
            return instance.root.find(self.tag)
        else:
            return instance.root


class StringAttributeDescriptor(TagDescriptor):
    """An instance attribute containing a string value
    represented by an XML attribute.
    """

    def __get__(self, instance, cls):
        instance.get()
        return instance.root.attrib[self.tag]


class StringListDescriptor(TagDescriptor):
    """An instance attribute containing a list of strings
    represented by multiple XML elements.
    """

    def __get__(self, instance, cls):
        instance.get()
        result = []
        for node in instance.root.findall(self.tag):
            result.append(node.text)
        return result


class StringDictionaryDescriptor(TagDescriptor):
    """An instance attribute containing a dictionary of string key/values
    represented by a hierarchical XML element.
    """

    def __get__(self, instance, cls):
        instance.get()
        result = dict()
        node = instance.root.find(self.tag)
        if node is not None:
            for node2 in node.getchildren():
                result[node2.tag] = node2.text
        return result


class IntegerDescriptor(StringDescriptor):
    """An instance attribute containing an integer value
    represented by an XMl element.
    """

    def __get__(self, instance, cls):
        instance.get()
        node = self.get_node(instance)
        if node is None:
            return None
        else:
            return int(node.text)


class BooleanDescriptor(StringDescriptor):
    """An instance attribute containing a boolean value
    represented by an XMl element.
    """

    def __get__(self, instance, cls):
        instance.get()
        node = self.get_node(instance)
        if node is None:
            return None
        else:
            return node.text.lower() == 'true'


class UdfDictionary(object):
    "Dictionary-like container of UDFs, optionally within a UDT."

    def __init__(self, instance, udt=False):
        self.instance = instance
        self._udt = udt
        self._update_elems()
        self._prepare_lookup()
        self.location=0

    def get_udt(self):
        if self._udt == True:
            return None
        else:
            return self._udt

    def set_udt(self, name):
        assert isinstance(name, basestring)
        if not self._udt:
            raise AttributeError('cannot set name for a UDF dictionary')
        self._udt = name
        elem = self.instance.root.find(nsmap('udf:type'))
        assert elem is not None
        elem.set('name', name)

    udt = property(get_udt, set_udt)

    def _update_elems(self):
        self._elems = []
        if self._udt:
            elem = self.instance.root.find(nsmap('udf:type'))
            if elem is not None:
                self._udt = elem.attrib['name']
                self._elems = elem.findall(nsmap('udf:field'))
        else:
            tag = nsmap('udf:field')
            for elem in self.instance.root.getchildren():
                if elem.tag == tag:
                    self._elems.append(elem)

    def _prepare_lookup(self):
        self._lookup = dict()
        for elem in self._elems:
            type = elem.attrib['type'].lower()
            value = elem.text
            if not value:
                value = None
            elif type == 'numeric':
                try:
                    value = int(value)
                except ValueError:
                    value = float(value)
            elif type == 'boolean':
                value = value == 'true'
            elif type == 'date':
                value = datetime.date(*time.strptime(value, "%Y-%m-%d")[:3])
            self._lookup[elem.attrib['name']] = value

    def __contains__(self,key):
        try:
            self._lookup[key]
        except KeyError:
            return False
        return True

    def __getitem__(self, key):
        return self._lookup[key]

    def __setitem__(self, key, value):
        self._lookup[key] = value
        for node in self._elems:
            if node.attrib['name'] != key: continue
            vtype = node.attrib['type'].lower()

            if value is None:
                pass
            elif vtype == 'string':
                if not isinstance(value, basestring):
                    raise TypeError('String UDF requires str or unicode value')
            elif vtype == 'str':
                if not isinstance(value, basestring):
                    raise TypeError('String UDF requires str or unicode value')
            elif vtype == 'text':
                if not isinstance(value, basestring):
                    raise TypeError('Text UDF requires str or unicode value')
            elif vtype == 'numeric':
                if not isinstance(value, (int, float)):
                    raise TypeError('Numeric UDF requires int or float value')
                value = str(value)
            elif vtype == 'boolean':
                if not isinstance(value, bool):
                    raise TypeError('Boolean UDF requires bool value')
                value = value and 'True' or 'False'
            elif vtype == 'date':
                if not isinstance(value, datetime.date): # Too restrictive?
                    raise TypeError('Date UDF requires datetime.date value')
                value = str(value)
            elif vtype == 'uri':
                if not isinstance(value, basestring):
                    raise TypeError('URI UDF requires str or punycode (unicode) value')
                value = str(value)
            else:
                raise NotImplemented("UDF type '%s'" % vtype)
            if not isinstance(value, unicode):
                value = unicode(value, 'UTF-8')
            node.text = value
            break
        else:                           # Create new entry; heuristics for type
            if isinstance(value, basestring):
                vtype = '\n' in value and 'Text' or 'String'
            elif isinstance(value, bool):
                vtype = 'Boolean'
                value = value and 'True' or 'False'
            elif isinstance(value, (int, float)):
                vtype = 'Numeric'
            elif isinstance(value, datetime.date):
                vtype = 'Date'
                value = str(value)
            else:
                raise NotImplementedError("Cannot handle value of type '%s'"
                                          " for UDF" % type(value))
            if self._udt:
                root = self.instance.root.find(nsmap('udf:type'))
            else:
                root = self.instance.root
            elem = ElementTree.SubElement(root,
                                          nsmap('udf:field'),
                                          type=vtype,
                                          name=key)
            if not isinstance(value, unicode):
                value = unicode(str(value), 'UTF-8')
            elem.text = value

    def __delitem__(self, key):
        del self._lookup[key]
        for node in self._elems:
            if node.attrib['name'] == key:
                self.instance.root.remove(node)
                break

    def items(self):
        return self._lookup.items()

    def clear(self):
        for elem in self._elems:
            self.instance.root.remove(elem)
        self._update_elems()

    def __iter__(self):
        return self

    def next(self):
        try:
            ret=self._lookup.keys()[self.location]
        except IndexError:
            raise StopIteration()
        self.location = self.location + 1
        return ret

    def get(self, key, default=None):
        return self._lookup.get(key, default)



class UdfDictionaryDescriptor(BaseDescriptor):
    """An instance attribute containing a dictionary of UDF values
    represented by multiple XML elements.
    """

    _UDT = False

    def __get__(self, instance, cls):
    	instance.get()
   	self.value = UdfDictionary(instance, udt=self._UDT)
   	return self.value

class UdtDictionaryDescriptor(UdfDictionaryDescriptor):
    """An instance attribute containing a dictionary of UDF values
    in a UDT represented by multiple XML elements.
    """

    _UDT = True


class PlacementDictionaryDescriptor(TagDescriptor):
    """An instance attribute containing a dictionary of locations
    keys and artifact values represented by multiple XML elements.
    """

    def __get__(self, instance, cls):
    	instance.get()
      	self.value = dict()
      	for node in instance.root.findall(self.tag):
            key = node.find('value').text
            self.value[key] = Artifact(instance.lims,uri=node.attrib['uri'])
       	return self.value


class ExternalidListDescriptor(BaseDescriptor):
    """An instance attribute yielding a list of tuples (id, uri) for
    external identifiers represented by multiple XML elements.
    """

    def __get__(self, instance, cls):
        instance.get()
        result = []
        for node in instance.root.findall(nsmap('ri:externalid')):
            result.append((node.attrib.get('id'), node.attrib.get('uri')))
        return result


class EntityDescriptor(TagDescriptor):
    "An instance attribute referencing another entity instance."

    def __init__(self, tag, klass):
        super(EntityDescriptor, self).__init__(tag)
        self.klass = klass

    def __get__(self, instance, cls):
        instance.get()
        node = instance.root.find(self.tag)
        if node is None:
            return None
        else:
            return self.klass(instance.lims, uri=node.attrib['uri'])


class EntityListDescriptor(EntityDescriptor):
    """An instance attribute yielding a list of entity instances
    represented by multiple XML elements.
    """

    def __get__(self, instance, cls):
        instance.get()
        result = []
        for node in instance.root.findall(self.tag):
            result.append(self.klass(instance.lims, uri=node.attrib['uri']))

        return result

class NestedAttributeListDescriptor(StringAttributeDescriptor):
    """An instance yielding a list of dictionnaries of attributes
       for a nested xml list of XML elements"""
    def __init__(self, tag, *args):
        super(StringAttributeDescriptor, self).__init__(tag)
        self.tag      = tag
        self.rootkeys = args

    def __get__(self, instance, cls):
        instance.get()
        result = []
        rootnode=instance.root
        for rootkey in self.rootkeys:
            rootnode=rootnode.find(rootkey)
        for node in rootnode.findall(self.tag):
            result.append(node.attrib)
        return result

class NestedStringListDescriptor(StringListDescriptor):
    """An instance yielding a list of strings
        for a nested list of xml elements"""
    def __init__(self, tag, *args):
        super(StringListDescriptor, self).__init__(tag)
        self.tag      = tag
        self.rootkeys = args

    def __get__(self, instance, cls):
        instance.get()
        result = []
        rootnode=instance.root
        for rootkey in self.rootkeys:
            rootnode=rootnode.find(rootkey)
        for node in rootnode.findall(self.tag):
            result.append(node.text)
        return result

class NestedEntityListDescriptor(EntityListDescriptor):
    """same as EntityListDescriptor, but works on nested elements"""

    def __init__(self, tag, klass, *args):
        super(EntityListDescriptor, self).__init__(tag, klass)
        self.klass    = klass
        self.tag      = tag
        self.rootkeys = args

    def __get__(self, instance, cls):
        instance.get()
        result = []
        rootnode=instance.root
        for rootkey in self.rootkeys:
            rootnode=rootnode.find(rootkey)
        for node in rootnode.findall(self.tag):
            result.append(self.klass(instance.lims, uri=node.attrib['uri']))
        return result
class DimensionDescriptor(TagDescriptor):
    """An instance attribute containing a dictionary specifying
    the properties of a dimension of a container type.
    """

    def __get__(self, instance, cls):
        instance.get()
        node = instance.root.find(self.tag)
        return dict(is_alpha = node.find('is-alpha').text.lower() == 'true',
                    offset = int(node.find('offset').text),
                    size = int(node.find('size').text))


class LocationDescriptor(TagDescriptor):
    """An instance attribute containing a tuple (container, value)
    specifying the location of an analyte in a container.
    """

    def __get__(self, instance, cls):
        instance.get()
        node = instance.root.find(self.tag)
        uri = node.find('container').attrib['uri']
        return Container(instance.lims, uri=uri), node.find('value').text

class ReagentLabelList(BaseDescriptor):
    """An instance attribute yielding a list of reagent labels"""
    def __get__(self, instance, cls):
	instance.get()
	self.value = []
	for node in instance.root.findall('reagent-label'):
	    try:
	    	self.value.append(node.attrib['name']) 
	    except:
		pass
	return self.value

class InputOutputMapList(BaseDescriptor):
    """An instance attribute yielding a list of tuples (input, output)
    where each item is a dictionary, representing the input/output
    maps of a Process instance.
    """

    def __get__(self, instance, cls):
        instance.get()
        self.value = []
        for node in instance.root.findall('input-output-map'):
            input = self.get_dict(instance.lims, node.find('input'))
            output = self.get_dict(instance.lims, node.find('output'))
            self.value.append((input, output))
        return self.value

    def get_dict(self, lims, node):
        if node is None: return None
        result = dict()
        for key in ['limsid', 'output-type', 'output-generation-type']:
            try:
                result[key] = node.attrib[key]
            except KeyError:
                pass
            for uri in ['uri', 'post-process-uri']:
                try:
                    result[uri] = Artifact(lims, uri=node.attrib[uri])
                except KeyError:
                    pass
        node = node.find('parent-process')
        if node is not None:
            result['parent-process'] = Process(lims, node.attrib['uri'])
        return result


class Entity(object):
    "Base class for the entities in the LIMS database."

    _TAG = None
    _URI = None

    def __new__(cls, lims, uri=None, id=None):
        if not uri:
            if not id:
                raise ValueError("Entity uri and id can't be both None")
            else:
                uri = lims.get_uri(cls._URI, id)

        try:
            return lims.cache[uri]
        except KeyError:
            return object.__new__(cls)

    def __init__(self, lims, uri=None, id=None):
        assert uri or id
        if hasattr(self, 'lims'): return
        if not uri:
            uri = lims.get_uri(self._URI, id)
        lims.cache[uri] = self
        self.lims = lims
        self._uri = uri
        self.root = None

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.id)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.uri)

    @property
    def uri(self):
        try:
            return self._uri
        except:
            return self._URI

    @property
    def id(self):
        "Return the LIMS id; obtained from the URI."
        parts = urlparse.urlsplit(self.uri)
        return parts.path.split('/')[-1]

    def get(self, force=False):
        "Get the XML data for this instance."
        if not force and self.root is not None: return
        self.root = self.lims.get(self.uri)

    def put(self):
        "Save this instance by doing PUT of its serialized XML."
        data = self.lims.tostring(ElementTree.ElementTree(self.root))
        self.lims.put(self.uri, data)


class Lab(Entity):
    "Lab; container of researchers."

    _URI = 'labs'

    name             = StringDescriptor('name')
    billing_address  = StringDictionaryDescriptor('billing-address')
    shipping_address = StringDictionaryDescriptor('shipping-address')
    udf              = UdfDictionaryDescriptor()
    udt              = UdtDictionaryDescriptor()
    externalids      = ExternalidListDescriptor()
    website          = StringDescriptor('website')


class Researcher(Entity):
    "Person; client scientist or lab personnel. Associated with a lab."

    _URI = 'researchers'

    first_name  = StringDescriptor('first-name')
    last_name   = StringDescriptor('last-name')
    phone       = StringDescriptor('phone')
    fax         = StringDescriptor('fax')
    email       = StringDescriptor('email')
    initials    = StringDescriptor('initials')
    lab         = EntityDescriptor('lab', Lab)
    udf         = UdfDictionaryDescriptor()
    udt         = UdtDictionaryDescriptor()
    externalids = ExternalidListDescriptor()
    # credentials XXX

    @property
    def name(self):
        return u"%s %s" % (self.first_name, self.last_name)

class Reagent_label(Entity):
    """Reagent label element"""
    reagent_label = StringDescriptor('reagent-label')

class Note(Entity):
    "Note attached to a project or a sample."

    content = StringDescriptor(None)    # root element


class File(Entity):
    "File attached to a project or a sample."

    attached_to       = StringDescriptor('attached-to')
    content_location  = StringDescriptor('content-location')
    original_location = StringDescriptor('original-location')
    is_published      = BooleanDescriptor('is-published')


class Project(Entity):
    "Project concerning a number of samples; associated with a researcher."

    _URI = 'projects'

    name          = StringDescriptor('name')
    open_date     = StringDescriptor('open-date')
    close_date    = StringDescriptor('close-date')
    invoice_date  = StringDescriptor('invoice-date')
    researcher    = EntityDescriptor('researcher', Researcher)
    udf           = UdfDictionaryDescriptor()
    udt           = UdtDictionaryDescriptor()
    files         = EntityListDescriptor(nsmap('file:file'), File)
    externalids   = ExternalidListDescriptor()
    # permissions XXX


class Sample(Entity):
    "Customer's sample to be analyzed; associated with a project."

    _URI = 'samples'

    name           = StringDescriptor('name')
    date_received  = StringDescriptor('date-received')
    date_completed = StringDescriptor('date-completed')
    project        = EntityDescriptor('project', Project)
    submitter      = EntityDescriptor('submitter', Researcher)
    # artifact: defined below
    udf            = UdfDictionaryDescriptor()
    udt            = UdtDictionaryDescriptor()
    notes          = EntityListDescriptor('note', Note)
    files          = EntityListDescriptor(nsmap('file:file'), File)
    externalids    = ExternalidListDescriptor()
    # biosource XXX


class Containertype(Entity):
    "Type of container for analyte artifacts."

    _TAG = 'container-type'
    _URI = 'containertypes'

    name              = StringAttributeDescriptor('name')
    calibrant_wells   = StringListDescriptor('calibrant-well')
    unavailable_wells = StringListDescriptor('unavailable-well')
    x_dimension       = DimensionDescriptor('x-dimension')
    y_dimension       = DimensionDescriptor('y-dimension')


class Container(Entity):
    "Container for analyte artifacts."

    _URI = 'containers'

    name           = StringDescriptor('name')
    type           = EntityDescriptor('type', Containertype)
    occupied_wells = IntegerDescriptor('occupied-wells')
    placements     = PlacementDictionaryDescriptor('placement')
    udf            = UdfDictionaryDescriptor()
    udt            = UdtDictionaryDescriptor()
    state          = StringDescriptor('state')

    def get_placements(self):
        """Get the dictionary of locations and artifacts
        using the more efficient batch call."""
        result = self.placements.copy()
        self.lims.get_batch(result.values())
        return result


class Processtype(Entity):

    _TAG = 'process-type'
    _URI = 'processtypes'

    name              = StringAttributeDescriptor('name')
    # XXX

class Udfconfig(Entity):
    "Instance of field type (cnf namespace)."
    _URI = 'configuration/udfs'

    name = StringDescriptor('name')
    attach_to_name = StringDescriptor('attach-to-name')
    attach_to_category = StringDescriptor('attach-to-category')


class Process(Entity):
    "Process (instance of Processtype) executed producing ouputs from inputs."

    _URI = 'processes'

    type          = EntityDescriptor('type', Processtype)
    date_run      = StringDescriptor('date-run')
    technician    = EntityDescriptor('technician', Researcher)
    protocol_name = StringDescriptor('protocol-name')
    input_output_maps = InputOutputMapList()
    udf            = UdfDictionaryDescriptor()
    udt            = UdtDictionaryDescriptor()
    files          = EntityListDescriptor(nsmap('file:file'), File)
    # instrument XXX
    # process_parameters XXX

    def outputs_per_input(self, inart, ResultFile = False, SharedResultFile = False,  Analyte = False):
        """Getting all the output artifacts related to a particual input artifact"""
        
        inouts = filter(lambda io: io[0]['limsid'] == inart, self.input_output_maps)
        if ResultFile:
            inouts = filter(lambda io: io[1]['output-type'] == 'ResultFile', inouts)
        elif SharedResultFile:
            inouts = filter(lambda io: io[1]['output-type'] == 'SharedResultFile', inouts)
        elif Analyte:
            inouts = filter(lambda io: io[1]['output-type'] == 'Analyte', inouts)
        outs = map(lambda io: io[1]['uri'], inouts)
        return outs

    def input_per_sample(self, sample):
        """gettiung all the input artifacts dereved from the specifyed sample"""
        ins_all = self.all_inputs()
        ins = []
        for inp in ins_all:
            for samp in inp.samples:
                if samp.name == sample and inp not in ins:
                    ins.append(inp)
        return ins
    
    def all_inputs(self,unique=True, resolve=False):
        """Retrieving all input artifacts from input_output_maps
        if unique is true, no duplicates are returned.
        """
        #if the process has no input, that is not standard and we want to know about it
        try:
            ids = [io[0]['limsid'] for io in self.input_output_maps]
        except TypeError:
            logger.error("Process ",self," has no input artifacts")
            raise TypeError
        if unique:
            ids = list(frozenset(ids))
        if resolve:
            return self.lims.get_batch([Artifact(self.lims,id=id) for id in ids if id is not None])
        else:
            return [Artifact(self.lims,id=id) for id in ids if id is not None]

    def all_outputs(self,unique=True, resolve=False):
        """Retrieving all output artifacts from input_output_maps
        if unique is true, no duplicates are returned.
        """
        #Given how ids is structured, io[1] might be None : some process don't have an output.
        ids = [io[1]['limsid'] for io in self.input_output_maps if io[1] is not None]
        if unique:
            ids = list(frozenset(ids))
        if resolve:
            return  self.lims.get_batch([Artifact(self.lims,id=id) for id in ids if id is not None])
        else:
            return  [Artifact(self.lims,id=id) for id in ids if id is not None]

    def shared_result_files(self):
        """Retreve all resultfiles of output-generation-type PerAllInputs."""
        artifacts = self.all_outputs(unique=True)
        return filter(lambda a: a.output_type == 'SharedResultFile', artifacts)

    def result_files(self):
        """Retreve all resultfiles of output-generation-type perInput."""
        artifacts = self.all_outputs(unique=True)
        return filter(lambda a: a.output_type == 'ResultFile', artifacts)

    def analytes(self):
        """Retreving the output Analytes of the process, if existing. 
        If the process is not producing any output analytes, the input 
        analytes are returned. Input/Output is returned as a information string.
        Makes aggregate processes and normal processes look the same."""
        info = 'Output'
        artifacts = self.all_outputs(unique=True)
        analytes = filter(lambda a: a.type == 'Analyte', artifacts)
        if len(analytes) == 0:
            artifacts = self.all_inputs(unique=True)
            analytes = filter(lambda a: a.type == 'Analyte', artifacts)
            info = 'Input'
        return analytes, info

    def parent_processes(self):
        """Retrieving all parent processes through the input artifacts"""
        return map(lambda i_a: i_a.parent_process, self.all_inputs(unique=True))

    def output_containers(self):
        """Retrieve all unique output containers"""
        cs = []
        for o_a in self.all_outputs(unique=True):
            if o_a.container:
                cs.append(o_a.container)
        return list(frozenset(cs))

class Artifact(Entity):
    "Any process input or output; analyte or file."

    _URI = 'artifacts'

    name           = StringDescriptor('name')
    type           = StringDescriptor('type')
    output_type    = StringDescriptor('output-type')
    parent_process = EntityDescriptor('parent-process', Process)
    volume         = StringDescriptor('volume')
    concentration  = StringDescriptor('concentration')
    qc_flag        = StringDescriptor('qc-flag')
    location       = LocationDescriptor('location')
    working_flag   = BooleanDescriptor('working-flag')
    samples        = EntityListDescriptor('sample', Sample)
    udf            = UdfDictionaryDescriptor()
    files          = EntityListDescriptor(nsmap('file:file'), File)
    reagent_labels = ReagentLabelList()
    # artifact_flags XXX
    # artifact_groups XXX

    def input_artifact_list(self):
        """Returns the input artifact ids of the parrent process."""
        input_artifact_list=[]
        try:
            for tuple in self.parent_process.input_output_maps:
                if tuple[1]['limsid'] == self.id:
                    input_artifact_list.append(tuple[0]['uri'])#['limsid'])
        except:
            pass
        return input_artifact_list

    def get_state(self):
        "Parse out the state value from the URI."
        parts = urlparse.urlparse(self.uri)
        params = urlparse.parse_qs(parts.query)
        try:
            return params['state'][0]
        except (KeyError, IndexError):
            return None

    @property
    def container(self):
        "The container where the artifact is located, or None"
        try:
            return self.location[0]
        except:
            return None

    def stateless(self):
        "returns the artefact independently of it's state"
        parts = urlparse.urlparse(self.uri)
        if 'state' in parts[4]:
            stateless_uri=urlparse.urlunparse([parts[0],parts[1], parts[2], parts[3], '',''])
            return Artifact(self.lims, uri=stateless_uri)
        else:
            return self

    # XXX set_state ?
    state = property(get_state)
    stateless = property(stateless) 

class StepActions(Entity):
    """Small hack to be able to query the actions subentity of
    the Step entity. Right now, only the escalation is parsed."""

    def __init__(self, lims, uri=None, id=None):
        super(StepActions, self).__init__(lims,uri,id)
        self.escalation={}
        self.lims=lims
        self.root=self.lims.get(self.uri)
        for node in self.root.findall('escalation'):
            self.escalation['artifacts']=[]
            self.escalation['author']=Researcher(lims,uri=node.find('request').find('author').attrib.get('uri'))
            self.escalation['request']=uri=node.find('request').find('comment').text
            if node.find('review') is not None: #recommended by the Etree doc
                self.escalation['status']='Reviewed'
                self.escalation['reviewer']= Researcher(lims,uri=node.find('review').find('author').attrib.get('uri'))
                self.escalation['answer']=uri=node.find('review').find('comment').text
            else:
                self.escalation['status']='Pending'

            for node2 in node.findall('escalated-artifacts'):
                art= lims.get_batch([Artifact(lims,uri=ch.attrib.get('uri')) for ch in node2])
                self.escalation['artifacts'].extend(art)


class Step(Entity):
    "Step, as defined by the genologics API."

    _URI = 'steps'

    def __init__(self, lims, uri=None, id=None):
        super(Step, self).__init__(lims,uri,id)
        assert self.uri is not None
        actionsuri="{0}/actions".format(self.uri)
        self.actions= StepActions(lims,uri=actionsuri)


    #placements         = EntityDescriptor('placements', StepPlacements)
    #program_status     = EntityDescriptor('program-status',StepProgramStatus)
    #details            = EntityListDescriptor(nsmap('file:file'), StepDetails)

class ProtocolStep(Entity):
    """Steps key in the Protocol object"""

    _TAG='step'

    name                = StringAttributeDescriptor("name")
    type                = EntityDescriptor('type', Processtype)
    permittedcontainers = NestedStringListDescriptor('container-type', 'container-types')
    queue_fields        = NestedAttributeListDescriptor('queue-field', 'queue-fields')
    step_fields         = NestedAttributeListDescriptor('step-field', 'step-fields')
    sample_fields       = NestedAttributeListDescriptor('sample-field', 'sample-fields')
    step_properties     = NestedAttributeListDescriptor('step_property', 'step_properties')
    epp_triggers        = NestedAttributeListDescriptor('epp_trigger', 'epp_triggers')


class Protocol(Entity):
    """Protocol, holding ProtocolSteps and protocol-properties"""
    _URI='configuration/protocols'
    _TAG='protocol'

    steps       = NestedEntityListDescriptor('step', ProtocolStep, 'steps')
    properties  = NestedAttributeListDescriptor('protocol-property', 'protocol-properties')


class Stage(Entity):
    """Holds Protocol/Workflow"""
    protocol = EntityDescriptor('protocol', Protocol)

class Workflow(Entity):
    """ Workflow, introduced in 3.5"""
    _URI="configuration/workflows"
    _TAG="workflow"
    
    name = StringAttributeDescriptor("name")
    protocols = NestedEntityListDescriptor('protocol', Protocol, 'protocols')
    stages    = EntityListDescriptor('stage', Stage)

class ReagentType(Entity):
    """Reagent Type, usually, indexes for sequencing"""
    _URI="reagenttypes"
    _TAG="reagent-type"

    category=StringDescriptor('reagent-category')

    def __init__(self, lims, uri=None, id=None):
        super(ReagentType, self).__init__(lims,uri,id)
        assert self.uri is not None
        self.root=lims.get(self.uri)
        self.sequence=None
        for t in self.root.findall('special-type'):
            if t.attrib.get("name") == "Index":
                for child in t.findall("attribute"):
                    if child.attrib.get("name") == "Sequence":
                        self.sequence=child.attrib.get("value")

Sample.artifact          = EntityDescriptor('artifact', Artifact)
StepActions.step         = EntityDescriptor('step', Step)
Stage.workflow            = EntityDescriptor('workflow', Workflow)
Artifact.workflow_stages = NestedEntityListDescriptor('workflow-stage', Stage, 'workflow-stages')
Step.configuration      = EntityDescriptor('configuration', ProtocolStep)





#!/usr/bin/env python
from __future__ import print_function
import logging
import os
import sys

from argparse import ArgumentParser
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import *
from genologics.epp import attach_file
DESC="""EPP used to create csv files for the nanoprep robot"""


def read_log(lims, pid, logfile):
    logger=setupLog(logfile)
    pro=Process(lims, id=pid)
    f=None
    for out in pro.all_outputs():
        if out.type == "ResultFile" and out.name == "NeoPrep Output Log File":
            try:
                fid=out.files[0].id
            except IndexError:
                logger.error("Can't find the machine log file")
                print("Cannot find the NeoPrep Output Log File", file=sys.stderr)
                exit(2)

            file_contents=lims.get_file_contents(id=fid)
            logger.info("Found the machine log file")

    if file_contents:
        data={}
        read=False
        #default values
        ist_size_idx=5
        sample_idx=2
        conc_idx=6
        norm_idx=7
        stat_idx=8
        logger.info("Reading the file")
        for line in file_contents.split('\n') :
            #This does something that is close from csv.dictreader, but the file is FUBAR
            if not line.rstrip():
                read=False
            if read:
                if "Start Well" in line:
                    #Header row
                    #identify which column goes with which index
                    elements=line.split('\t')
                    for idx, el in enumerate(elements):
                        if el == "Name":
                            sample_idx=idx
                        elif el == "Quant":
                            conc_idx=idx
                        elif el == "Norm":
                            norm_idx=idx
                        elif el == "Status":
                            stat_idx=idx
                        elif el == "Insert Size":
                            ist_size_idx=idx
                else:
                    elements=line.split('\t')
                    #data rows
                    data[elements[sample_idx]]={}
                    data[elements[sample_idx]]['conc']=elements[conc_idx]
                    data[elements[sample_idx]]['norm']=elements[norm_idx]
                    data[elements[sample_idx]]['stat']=elements[stat_idx]
                    data[elements[sample_idx]]['ist_size']=elements[ist_size_idx]

            if "[Sample Information]" in line:
                read=True
        logger.info("obtained data for samples {0}".format(data.keys()))

    for inp in pro.all_inputs():
        #save the data from the logfile to the lims artifacts
        if inp.name in data:
            if data[inp.name]['stat'] == 'Pass':
                inp.qc="PASSED"
            elif data[inp.name]['stat'] == 'Fail':
                inp.qc="FAILED"
            inp.udf['Molar Conc. (nM)']=float(data[inp.name]['conc'])
            inp.udf['Normalized conc. (nM)']=float(data[inp.name]['norm'])
            inp.udf['NeoPrep Machine QC']=data[inp.name]['stat']
            inp.udf['Size (bp)']=float(data[inp.name]['ist_size'])
            inp.put()
            logger.info("updated sample {0}".format(inp.name))

    for out in pro.all_outputs():
        #attach the epp log
        if out.name=="EPP Log":
            attach_file(os.path.join(os.getcwd(), logfile), out)


            



def setupLog(logfile):
        mainlog = logging.getLogger(__name__)
        mainlog.setLevel(level=logging.INFO)
        mfh = logging.FileHandler(logfile)
        mft = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        mfh.setFormatter(mft)
        mainlog.addHandler(mfh)
        return mainlog


if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', default="read_neoprep_log.log",
                        help='logfile for this epp')
    args = parser.parse_args()

    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()
    read_log(lims, args.pid, args.log)



#!/usr/bin/env python
DESC = """EPP script to copy user defined field 'Reference Genome' from project 
level to submitted sample level for the input artifacts of given process, 
in Clarity LIMS. Can be executed in the background or triggered by a user
 pressing a "blue button".

The script outputs a regular log file that contains regular execution 
information. 

Error handling:
If the udf 'Reference Genome' is blank or not defined for any of the input 
projects, the script will log this, and not perform any changes for that sample.


Written by Johannes Alneberg, Science for Life Laboratory, Stockholm, Sweden
""" 

from argparse import ArgumentParser

from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD

from genologics.entities import Process, Artifact
from genologics.epp import EppLogger

import logging
import sys

class Session(object):
    def __init__(self, process, udf, d_udf):
        self.process = process
        self.s_udf = udf    # Source udf
        self.d_udf = d_udf  # Destination udf

    def _sample_udf(self,sample):
        if self.d_udf in sample.udf:
            return sample.udf[self.d_udf]
        else:
            return "Undefined/Blank"

    def log_before_change(self, project, sample):
        logging.info(("Copying from project with id: {0} to sample with "
                      " id: {1}").format(project.id, sample.id))

    def log_after_change(self, project, saved_sample_udf):
        d = {'udf': self.d_udf,
             'su': saved_sample_udf,
             'nv': project.udf[self.s_udf]
             }

        logging.info("Updated Sample {udf} from {su} to {nv}.".format(**d))

    def copy_udf(self, project, sample):
        saved_sample_udf = self._sample_udf(sample)

        self.log_before_change(project, sample)

        n_udf = project.udf[self.s_udf] # New udf
        sample.udf[self.d_udf] = n_udf
        sample.put()
        self.log_after_change(project, saved_sample_udf)

    def copy_main(self, samples):
        for sample in samples:
            project = sample.project

            self.copy_udf(project, sample)



def all_projects_for_artifacts(artifacts):
    """ get all unique projects associated with a list of artifacts """
    projects = set()
    for artifact in artifacts:
        for sample in artifact.samples:
            projects.add(sample.project)
    return list(projects)

def check_udf_is_defined(projects, udf):
    """ Filter and Warn if udf is not defined for any of inputs. """
    filtered_projects = []
    incorrect_projects = []
    for project in projects:
        if (udf in project.udf):
            filtered_projects.append(project)
        else:
            logging.warning(("Found project with id {0} with {1} "
                             "undefined/blank, exiting").format(project.id, udf))
            incorrect_projects.append(project)
    return filtered_projects, incorrect_projects

def filter_samples(artifacts, projects):
    """ All samples belonging to a project in projects """
    projects = set(projects)
    return_samples = []
    samples = set()
    for artifact in artifacts:
        samples.update(artifact.samples)

    for sample in samples:
        if sample.project in projects:
            return_samples.append(sample)
        else:
            logging.warning(("Filtered out sample {0} belonging to project {1} "
                             "without udf defined").format(sample.name,
                                                           sample.project.name))
    return return_samples

def main(lims,args,epp_logger):
    pro = Process(lims,id = args.pid)
    source_udf = 'Reference genome'
    destination_udf = 'Reference Genome'

    artifacts = pro.all_inputs(unique=True)
    projects = all_projects_for_artifacts(artifacts)

    correct_projects, incorrect_udf = check_udf_is_defined(projects, source_udf)
    correct_samples = filter_samples(artifacts, correct_projects)

    session = Session(pro, source_udf, destination_udf)
    session.copy_main(correct_samples)

    if len(incorrect_udf) == 0:
        warning = "no projects"
    else:
        warning = "WARNING: skipped {0} project(s)".format(len(incorrect_udf))

    d = {'cs': len(correct_samples),
         'warning' : warning
    }

    abstract = ("Updated {cs} sample(s), {warning} with incorrect udf info.").format(**d)

    print >> sys.stderr, abstract # stderr will be logged and printed in GUI


if __name__ == "__main__":
    # Initialize parser with standard arguments and description

    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid',
                        help='Lims id for current Process')
    parser.add_argument('--log',
                        help=('File name for standard log file, '
                              ' for runtime information and problems.'))
    args = parser.parse_args()

    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(args.log,lims=lims, prepend=True) as epp_logger:
        main(lims, args, epp_logger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

DESC = """EPP script to copy 'concentration' and 'concentration unit' for each 
sample sample in the 'Qubit Result File' to the 'Concentration' and 'Conc. Units' 
fields of the output analytes of the process.

Warnings are generated to the user and stored in regular log file wich allso 
contains regular execution information in the folowing cases:

1) missing row names (samples) in file
2) duplicated row names (samples)
3) missing value (concentrations) 
4) values found but for some reason are not successfully copied:
 
Can be executed in the background or triggered by a user pressing a "blue button".

Written by Maya Brandi and Denis Moreno
"""

import os
import sys
import logging
import numpy as np
import codecs

from argparse import ArgumentParser
from requests import HTTPError
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import ReadResultFiles
from genologics.epp import set_field

import csv

def get_qbit_file(process):
    content = None
    for outart in process.all_outputs():
        #get the right output artifact
        if outart.type == 'ResultFile' and outart.name == 'Qubit Result File':
            try:
                fid = outart.files[0].id
                content = lims.get_file_contents(id=fid)
            except:
                raise(RuntimeError("Cannot access the tecan output file to read the concentrations."))
            break
    return content

def get_data(csv_content, log):
    read=False
    data={}
    text = csv_content.encode("utf-8")
    # Try to determine the format of the csv:
    dialect = csv.Sniffer().sniff(text)
    pf = csv.reader(text.splitlines(), dialect=dialect)
    #New Qubit
    #defaults
    sample_index=2
    conc_index=6
    conc_unit=7
    for row in pf:
        if 'Test Name' in row:
            #this is the header row
            sample_index=row.index('Test Name')
            conc_index=row.index('Original sample conc.')
            unit_index=conc_index+1
            read=True
        elif(read and row[sample_index]):
            #this is every other row
            if row[sample_index] in data:
                #Sample is duplicated, drop the key
                log.append("sample {0} has two rows in the Qubit CSV file. Please check the file manually.".format(row[sample_index]))
                del data[row[sample_index]]

            else:
                #normal procedure
                data[row[sample_index]]={}
                data[row[sample_index]]['concentration']=row[conc_index]
                data[row[sample_index]]['unit']=row[unit_index]
        elif (not read  and 'Sample' in row):
            sample_index=row.index('Sample')
            conc_index=row.index('Sample Concentration')
            unit_index=conc_index+1
            read=True
    return data

def convert_to_ng_ul(conc, unit):
    factor=float(1.0)#I really want a float
    units=unit.split('/')
    if units[0] == 'µg' or units[0] == 'ug':
        factor*=1000
    elif units[0] == 'mg':
        factor*=1000000
    if units[1] == 'mL':
        factor/=1000
    if units[1] == 'nL':
        factor*=1000

    return conc*factor


def get_qbit_csv_data(process):
    #samples missing from the qubit csv file
    missing_samples = 0
    low_conc = 0
    bad_format = 0
    #strings returned to the EPP user
    log = []
    # Get file contents by parsing lims artifacts
    file_content = get_qbit_file(process)
    #parse the qubit file and get the interesting data out
    data = get_data(file_content, log)

    if "Minimum required concentration (ng/ul)" in process.udf:
        min_conc=process.udf['Minimum required concentration (ng/ul)']
    else:
        min_conc=None
        log.append("Set 'Minimum required concentration (ng/ul)' to get qc-flags based on this threshold!")

    for target_file in process.result_files():
        conc=None
        new_conc=None
        file_sample=target_file.samples[0].name
        if file_sample in data:
            try:
                conc=float(data[file_sample]['concentration'])
            except ValueError:
                #concentration is not a float
                target_file.qc_flag = "FAILED"
                if data[file_sample]['concentration'] != 'Out of range':
                    #Out of range is a valid value, the others are not.
                    bad_format+=1

            else:
                new_conc=convert_to_ng_ul(conc, data[file_sample]['unit'])
                if new_conc is not None : 
                    target_file.udf['Concentration'] = new_conc
                    target_file.udf['Conc. Units'] = 'ng/ul'
                    if new_conc < min_conc:
                        target_file.qc_flag = "FAILED"
                        low_conc +=1
                    else:
                        target_file.qc_flag = "PASSED"

            #actually set the data
            target_file.put()
            set_field(target_file)
        else:
            missing_samples += 1
    if low_conc:
        log.append('{0}/{1} samples have low concentration.'.format(low_conc, len(process.result_files())))
    if missing_samples:
        log.append('{0}/{1} samples are missing in the Qubit Result File.'.format(missing_samples, len(process.result_files())))
    if bad_format:
        log.append('There are {0} badly formatted samples in Qubit Result File. Please fix these to get proper results.'.format(bad_format))

    print(''.join(log), file=sys.stderr)

def main(lims, pid, epp_logger):

    process = Process(lims,id = pid)
    get_qbit_csv_data(process)

def old_main(lims, pid, epp_logger):
    process = Process(lims,id = pid)
    sample_names = map(lambda a: a.name, process.analytes()[0])
    target_files = process.result_files()
    file_handler = ReadResultFiles(process)
    files = file_handler.shared_files['Qubit Result File']
    qubit_result_file = file_handler.format_file(files, 
                                                 name = 'Qubit Result File',
                                                 first_header = ['Test','Sample'],
                                                 find_keys = sample_names)
    missing_samples = 0
    low_conc = 0
    bad_formated = 0
    abstract = []
    udfs = dict(process.udf.items())
    if udfs.has_key("Minimum required concentration (ng/ul)"):
        min_conc = udfs["Minimum required concentration (ng/ul)"]
    else:
        min_conc = None
        abstract.append("Set 'Minimum required concentration (ng/ul)' to get qc-flaggs based on this treshold!")
    for target_file in target_files:
        sample = target_file.samples[0].name
        if qubit_result_file.has_key(sample):
            sample_mesurements = qubit_result_file[sample]
            if "Sample Concentration" in sample_mesurements.keys():
                conc, unit = sample_mesurements["Sample Concentration"]
                if conc == 'Out Of Range':
                    target_file.qc_flag = "FAILED"
                elif conc.replace('.','').isdigit():
                    conc = float(conc)
                    if unit == 'ng/mL':
                        conc = np.true_divide(conc, 1000)
                    if min_conc:
                        if conc < min_conc:
                            target_file.qc_flag = "FAILED"
                            low_conc +=1
                        else:
                            target_file.qc_flag = "PASSED"
                    target_file.udf['Concentration'] = conc
                    target_file.udf['Conc. Units'] = 'ng/ul'
                else:
                    bad_formated += 1
                set_field(target_file)
        else:
            missing_samples += 1

    if low_conc:
        abstract.append('{0}/{1} samples have low concentration.'.format(low_conc, len(target_files)))
    if missing_samples:
        abstract.append('{0}/{1} samples are missing in Qubit Result File.'.format(missing_samples, len(target_files)))
    if bad_formated:
        abstract.append('There are {0} badly formated samples in Qubit Result File. Please fix these to get proper results.'.format(bad_formated))

    print >> sys.stderr, ' '.join(abstract)

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = '24-38458', dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))

    args = parser.parse_args()

    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, epp_logger)




#!/usr/bin/env python
DESC = """EPP script for updating #Reads and %Q30 for quality filtered reads.
The script reads the new values from a csv file "Quality Filter" that first 
need to be generated by the quality filter script XXX and uploaded to the process.

Reads from:
    --files--
    "Quality Filter"            "shared result file" uploaded by user.   

Writes to:
    --Lims fields--
    "% Bases >=Q30"             udf of process artifacts (result file)
    "#Reads"                    udf of process artifacts (result file)

Logging:
    The script outputs a regular log file with regular execution information.

Written by Maya Brandi 
"""

import os
import sys
import logging
import csv

from argparse import ArgumentParser
from genologics.lims import Lims
from genologics.config import BASEURI, USERNAME, PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import set_field

class QualityFilter():
    def __init__(self, process):
        self.process = process
        self.flowcell_id = process.all_inputs()[0].container.name
        self.project_name = process.all_outputs()[0].samples[0].project.name
        self.result_files = process.result_files()
        self.source_file = None
        self.QF_from_file = {}
        self.missing_samps = []
        self.abstract = []
        self.abstract_ext = []
        self.nr_samps_updat = []

    def read_QF_file(self):
        """ QF file is read from the file msf system. Path hard coded."""
        file_path = ("/srv/mfs/QF/{0}/{1}.csv".format(self.project_name, self.flowcell_id))
        of = open(file_path ,'r')
        self.source_file = [row for row in csv.reader(of.read().splitlines())]
        of.close()

    def get_and_set_yield_and_Q30(self):
        self._format_file()
        input_pools = self.process.all_inputs()
        for pool in input_pools:
            self.nr_samps_updat = []
            self.missing_samps = []
            lane = pool.location[1][0] #getting lane number
            outarts_per_lane = self.process.outputs_per_input(
                                          pool.id, ResultFile = True)
            for target_file in outarts_per_lane:
                samp_name = target_file.samples[0].name
                self._set_udfs(samp_name, target_file, lane)
            if self.nr_samps_updat:
                self.abstract_ext.append("LANE {0} with {1} samples."
                    "".format(lane, str(len(set(self.nr_samps_updat)))))
            if self.missing_samps:
                self.abstract_ext.append("The following samples are missing in Quality "
                    "Filter file: {0}.".format(', '.join(self.missing_samps)))
        self._logging()

    def _format_file(self):
        keys = self.source_file[0]
        l_ind = keys.index('Lane')
        s_ind = keys.index('Sample')
        q_ind = keys.index('% Bases >=Q30')
        y_ind = keys.index('# Reads')
        for line in self.source_file[1:]:
            lane = line[l_ind]
            samp = line[s_ind]
            if not lane in self.QF_from_file.keys():
                self.QF_from_file[lane] = {}
            self.QF_from_file[lane][samp] = {'% Bases >=Q30' : line[q_ind],
                                                   '# Reads' : line[y_ind]}
  
    def _set_udfs(self, samp_name, target_file, lane):
        if lane in self.QF_from_file.keys():
            if samp_name in self.QF_from_file[lane].keys():
                s_inf = self.QF_from_file[lane][samp_name]
                target_file.udf['# Reads'] = int(s_inf['# Reads'])
                target_file.udf['% Bases >=Q30'] = float(s_inf['% Bases >=Q30'])
                self.nr_samps_updat.append(samp_name)
            else:
                self.missing_samps.append(samp_name)
        set_field(target_file)

    def _logging(self):
        if self.abstract_ext:
            self.abstract.append("Yield and Q30 uploaded to")
            self.abstract.extend(self.abstract_ext)
            print >> sys.stderr, ' '.join(self.abstract)
        else:
            sys.exit("Yield and Q30 not found in QF file for any of the avalible lanes.")


def main(lims, pid, epp_logger):
    process = Process(lims,id = pid)
    QF = QualityFilter(process)
    QF.read_QF_file()
    QF.get_and_set_yield_and_Q30()
    

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = None , dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))

    args = parser.parse_args()
    lims = Lims(BASEURI, USERNAME, PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, epp_logger)



#!/usr/bin/env python
DESC = """EPP script for Quant-iT mesurements to set QC flaggs and 
intensity check based on concentrations, Fluorescence intensity. 

Performance:
    1) compares udfs "Fluorescence intensity 1" and "Fluorescence intensity 2" 
    with the Saturation threshold of fluorescence intensity. If either of these 
    two udfs >= Saturation threshold of fluorescence intensity, assign 
    "Saturated" to the udf "Intensity check" and assign "Fail" to the sample. 
    Otherwise assign "OK" to the analyte "Intensity check".

    2) Copies the %CV values to the sample udf "%CV". 
    For a sample with duplicate measurements, "%CV" is calculated as:
        %CV = (SD of "Fluorescence intensity 1" and "Fluorescence intensity 2")/
        (Mean of "Fluorescence intensity 1" and ""Fluorescence intensity 2)

    3) If "%CV" >= "Allowed %CV of duplicates", assigning "Fail" to the QC flagg.

    4) For a sample with only one measurement, if it passes in step 2, a "Pass" 
    should be assigned to the QC flag. For a sample with duplicate measurements,
    if it passes both step 2 and step 4, a "Pass" is assigned to the QC flag.

Reads from:
    --Lims fields--
    "Saturation threshold of fluorescence intensity"    process udf 
    "Allowed %CV of duplicates"                         process udf
    "Fluorescence intensity 1"  udf of input analytes to the process
    "Fluorescence intensity 2"  udf of input analytes to the process

    --files--
    "Standards File (.txt)"     "shared result file" uploaded by user.   
    "Quant-iT Result File 1"    "shared result file" uploaded by user.
    "Quant-iT Result File 2"    "shared result file" uploaded by user. (optional)

Writes to:
    --Lims fields--
    "Intensity check"           udf of process artifacts (result file)
    "%CV"                       udf of process artifacts (result file)
    "QC"                        qc-flag of process artifacts (result file)

Logging:
    The script outputs a regular log file with regular execution information.

Written by Maya Brandi 
"""

import os
import sys
import logging
import numpy as np

from argparse import ArgumentParser
from requests import HTTPError
from genologics.lims import Lims
from genologics.config import BASEURI,USERNAME,PASSWORD
from genologics.entities import Process
from genologics.epp import EppLogger
from genologics.epp import set_field
from genologics.epp import ReadResultFiles

class QuantitQC():
    def __init__(self, process):
        self.result_files = process.result_files()
        self.udfs = dict(process.udf.items())
        self.required_udfs = set(["Allowed %CV of duplicates",
            "Saturation threshold of fluorescence intensity",
            "Minimum required concentration (ng/ul)"])
        self.abstract = []
        self.missing_udfs = []
        self.hig_CV_fract = 0
        self.saturated = 0
        self.low_conc = 0
        self.flour_int_missing = 0
        self.conc_missing = 0
        self.no_failed = 0

    def saturation_QC(self, result_file, udfs):
        treshold = self.udfs["Saturation threshold of fluorescence intensity"]
        allowed_dupl = self.udfs["Allowed %CV of duplicates"]
        fint_key1 = "Fluorescence intensity 2"
        fint_key2 = "Fluorescence intensity 1"
        fint_2 = udfs[fint_key2] if udfs.has_key(fint_key2) else None
        fint_1 = udfs[fint_key1] if udfs.has_key(fint_key1) else None
        if fint_1 or fint_2:
            qc_flag = "PASSED"
            if (fint_1 >= treshold) or (fint_2 >= treshold):
                result_file.udf["Intensity check"] = "Saturated"
                qc_flag = "FAILED"
                self.saturated +=1
            else:
                result_file.udf["Intensity check"] = "OK"
                if fint_1 and fint_2:
                    std = np.std([fint_1, fint_2])
                    mean = np.mean([fint_1, fint_2])
                    procent_CV = np.true_divide(std,mean)
                    result_file.udf["%CV"] = procent_CV
                    if procent_CV >= allowed_dupl:
                        qc_flag = "FAILED"
                        self.hig_CV_fract +=1
            return qc_flag
        else:
            self.flour_int_missing +=1
            return None

    def concentration_QC(self, result_file, result_file_udfs):
        min_conc = self.udfs["Minimum required concentration (ng/ul)"]
        if result_file_udfs.has_key('Concentration'):
            if result_file_udfs['Concentration'] < min_conc:
                self.low_conc +=1
                return "FAILED"
            else:
                return "PASSED"
        else:
            self.conc_missing +=1
            return None

    def assign_QC_flag(self):
        if self.required_udfs.issubset(self.udfs.keys()):
            for result_file in self.result_files:
                result_file_udfs = dict(result_file.udf.items())
                QC_conc = self.concentration_QC(result_file, result_file_udfs)
                QC_sat = self.saturation_QC(result_file, result_file_udfs)
                if QC_conc and QC_sat:
                    QC = QC_conc if QC_conc == QC_sat else "FAILED"
                    self.no_failed +=1 if QC == "FAILED" else 0
                    result_file.qc_flag = QC
                    set_field(result_file)
        else:
            self.missing_udfs = ', '.join(list(self.required_udfs))

def main(lims, pid, epp_logger):
    process = Process(lims,id = pid)
    QiT = QuantitQC(process)
    QiT.assign_QC_flag()
    if QiT.flour_int_missing:
        QiT.abstract.append("Fluorescence intensity is missing for {0} "
                                       "samples.".format(QiT.flour_int_missing))
    if QiT.missing_udfs:
        QiT.abstract.append("Could not set QC flags. Some of the following "
             "required udfs seems to be missing: {0}.".format(QiT.missing_udfs))
    else:
        QiT.abstract.append("{0} out of {1} samples failed "
                       "QC.".format(QiT.no_failed, len(process.result_files())))
    if QiT.saturated:
        QiT.abstract.append("{0} samples had saturated fluorescence "
                                             "intensity.".format(QiT.saturated))
    if QiT.hig_CV_fract:
        QiT.abstract.append("{0} samples had high %CV.".format(QiT.hig_CV_fract))
    if QiT.low_conc:
        QiT.abstract.append("{0} samples had low concentration.".format(
                                                                  QiT.low_conc))
    if QiT.conc_missing:
        QiT.abstract.append("Concentration is missing for {0} "
                                     "sample(s).".format(QiT.conc_missing))
    QiT.abstract = list(set(QiT.abstract))
    print >> sys.stderr, ' '.join(QiT.abstract)

if __name__ == "__main__":
    parser = ArgumentParser(description=DESC)
    parser.add_argument('--pid', default = None , dest = 'pid',
                        help='Lims id for current Process')
    parser.add_argument('--log', dest = 'log',
                        help=('File name for standard log file, '
                              'for runtime information and problems.'))

    args = parser.parse_args()
    lims = Lims(BASEURI,USERNAME,PASSWORD)
    lims.check_version()

    with EppLogger(log_file=args.log, lims=lims, prepend=True) as epp_logger:
        main(lims, args.pid, epp_logger)

